from nawah import __version__

from typing import Dict, Any, Optional

import argparse, os, sys, logging, json, subprocess, shutil, pkgutil, traceback

logger = logging.getLogger('nawah')


def packages_install(args: argparse.Namespace):
    app_path = os.path.realpath('.')
    sys.path.insert(0, app_path)

    with open(os.path.join(app_path, 'packages.json')) as f:
        packages = json.loads(f.read())

    for package_name in packages:
        logger.info(f'Attempting to install package \'{package_name}\'')
        _packages_add(
            package_name=package_name,
            source=packages[package_name]['source'],
            version=packages[package_name]['version'],
            auth=packages[package_name]['auth']
            if 'auth' in packages[package_name].keys()
            else None,
        )

    logger.info('Done installing all packages!')


# [DOC] The single underscore to indicate this is not being called directly by argsparser but through a proxy callable
def _packages_add(*, package_name: str, source: str, version: str, auth: Optional[str]):
    from . import _set_testing
    import json

    _set_testing(True)
    logger.info('Checking packages conflicts.')
    app_path = os.path.realpath(os.path.join('.'))
    packages_path = os.path.realpath(os.path.join('.', 'packages'))
    package_path = os.path.realpath(os.path.join('.', 'packages', package_name))

    with open(os.path.join(app_path, 'packages.json')) as f:
        packages = json.loads(f.read())

    api_level = '.'.join(__version__.split('.')[:2])

    if os.path.exists(package_path):
        logger.info(
            f'Package \'{package_name}\' already exists in app. Attempting to test compatibility with API Level and version.'
        )
        try:
            sys.path.insert(0, os.path.realpath('.'))
            sys.path.insert(0, packages_path)
            package = __import__(package_name)
        except:
            logger.error(
                'Failed to load package to test compatibility and version. Exiting.'
            )
            exit(1)

        if package.config.api_level != api_level:
            logger.error(
                f'App is using API Level {api_level}, but package is on API Level {package.config.api_level}. Exiting.'
            )
            exit(1)

        logger.info(
            f'Package is compatible with app API Level. Attempting to check version.'
        )

        if package.config.version != (
            package_version := packages[package_name]['version']
        ):
            logger.error(
                f'Package is on version {package.config.version}, but app requires version {package_version}. Exiting.'
            )
            exit(1)

        logger.info(f'Package is already installed with correct version.')

        _set_testing(False)

        return

    authed_source = source
    try:
        if auth:
            logger.info(
                f'Detected \'auth\' configuration for source. Attempting to authorise source.'
            )
            auth_username, auth_password = auth.split(':')
            if auth_username.startswith('__env.'):
                logger.info(
                    f'Detected environement variable for \'auth_username\' configuration. Setting it.'
                )
                auth_username = os.environ[auth_username.replace('__env.', '')]
            if auth_password.startswith('__env.'):
                logger.info(
                    f'Detected environement variable for \'auth_password\' configuration. Setting it.'
                )
                auth_password = os.environ[auth_password.replace('__env.', '')]
            authed_source = (
                authed_source.split('://', 1)[0]
                + '://'
                + auth_username
                + ':'
                + auth_password
                + '@'
                + authed_source.split('://', 1)[1]
            )
            logger.info(f'Processed \'auth\' configuration successfully.')
    except:
        logger.error(f'Failed to process \'auth\' configuration. Exiting.')
        exit(1)

    logger.info(
        f'Attempting to add package \'{package_name}\' from source \'{source}\'.'
    )

    pip_command = [
        sys.executable,
        '-m',
        'pip',
        'install',
        '--no-deps',
        '--target',
        packages_path,
        '--extra-index-url',
        authed_source,
        package_name if version == 'latest' else f'{package_name}=={version}',
    ]

    pip_call = subprocess.call(pip_command)

    if pip_call != 0:
        logger.error(
            'Last \'pip\' call failed. Check console for more details. Exiting.'
        )
        exit(1)

    logger.info(
        f'Package installed. Attempting to test compatibility with API Level {api_level}.'
    )

    try:
        sys.path.insert(0, os.path.realpath('.'))
        sys.path.insert(0, packages_path)
        package = __import__(package_name)
    except Exception as e:
        logger.error(traceback.format_exc())
        logger.error('Failed to load package to test compatibility. Exiting.')
        exit(1)

    if package.config.api_level != api_level:
        logger.error(
            f'App is using API Level {api_level}, but package is on API Level {package.config.api_level}.'
        )
        _packages_rm(package_name=package_name, confirm=False)

    logger.info(
        f'Package is compatible with app API Level. Attempting to install package requirements.'
    )

    requirements_path = os.path.join(package_path, 'requirements.txt')

    pip_command = [
        sys.executable,
        '-m',
        'pip',
        'install',
        '--user',
        '-r',
        requirements_path,
    ]

    pip_call = subprocess.call(pip_command)

    if pip_call != 0:
        logger.error('Last \'pip\' call failed. Check console for more details.')

    logger.info('Attempting to update \'packages.json\'.')

    with open(os.path.realpath(os.path.join('.', 'packages.json')), 'r') as f:
        packages = json.loads(f.read())
        packages[package_name] = {'version': package.config.version, 'source': source}
        if auth:
            packages[package_name]['auth'] = auth

    with open(os.path.realpath(os.path.join('.', 'packages.json')), 'w') as f:
        f.write(json.dumps(packages))

    logger.info('Successfully updated \'packages.json\'.')
    logger.info(
        'Remember to check package docs for any \'vars\' you are required to add to your app.'
    )

    _set_testing(False)


def _packages_rm(*, package_name: str, confirm: bool = True):
    from . import _set_testing

    if confirm:
        confirmation = input(
            f'Are you sure you want to remove package \'{package_name}\'? [yN] '
        )
        if not len(confirmation) or confirmation.lower()[0] != 'y':
            logger.info(f'Cancelled removing package \'{package_name}\'. Exiting.')
            exit(0)

    _set_testing(True)

    try:
        packages_path = os.path.realpath(os.path.join('.', 'packages'))
        package_path = os.path.realpath(os.path.join('.', 'packages', package_name))
        sys.path.insert(0, packages_path)
        package = __import__(package_name)
    except:
        logger.error(
            'Failed to load package to remove it Confirm package name is correct. Exiting.'
        )
        exit(1)

    logger.warning(f'Removing package \'{package_name}\'...')
    # [DOC] Handle removing files for Windows/NT when shutil.rmtree fails
    # [REF] https://stackoverflow.com/a/28476782/2393762
    def errorRemoveReadonly(func, path, exc):
        excvalue = exc[1]
        if func in (os.rmdir, os.remove) and excvalue.errno == errno.EACCES:
            os.chmod(path, stat.S_IRWXU | stat.S_IRWXG | stat.S_IRWXO)
            func(path)

    shutil.rmtree(package_path, ignore_errors=False, onerror=errorRemoveReadonly)
    shutil.rmtree(
        os.path.join(
            packages_path, f'{package_name}-{package.config.version}.dist-info'
        ),
        ignore_errors=False,
        onerror=errorRemoveReadonly,
    )
    logger.info(f'Package \'{package_name}\' removed. Exiting.')
    exit(1)


def packages_audit(args: argparse.Namespace):
    import json

    logger.info(
        'Attempting to audit packages in \'nawah_pacakges.py\' against packages added to \'packages\' folder.'
    )

    app_path = os.path.realpath('.')
    packages_path = os.path.realpath(os.path.join('.', 'packages'))
    sys.path.insert(0, app_path)
    sys.path.append(packages_path)
    import nawah_app

    with open(os.path.join(app_path, 'packages.json')) as f:
        req_packages: Dict[str, Any] = json.loads(f.read())

    logger.info('Packages in \'nawah_pacakges.py\' are:')
    for package_name in req_packages.keys():
        logger.info(f'- {package_name}: {req_packages[package_name]}')

    logger.info('Attempting to check packages added in \'packages\'.')
    added_packages = {}
    for _, pkgname, ispkg in pkgutil.iter_modules([packages_path]):
        if not ispkg:
            continue
        logger.info(f'- Importing package: {pkgname}')
        package = __import__(pkgname, fromlist='*')
        added_packages[pkgname] = package.config.version
        logger.info(f'- Imported package with version: {package.config.version}')

    logger.info('Attempting to audit missing, extra, version-mismatching packages.')

    missing_packages = [
        package for package in req_packages if package not in added_packages.keys()
    ]
    if missing_packages:
        logger.info(
            'Following packages are required by your app but not added to \'packages\':'
        )
        for package in missing_packages:
            logger.info(f'- {package}')
    else:
        logger.info('Great! Your app is not missing any of the required package.')

    extra_packages = [
        package
        for package in added_packages
        if package not in req_packages.keys()
        and package != nawah_app.config.default_package
    ]
    if extra_packages:
        logger.info(
            'Following packages are added to \'packages\' but not required by your app:'
        )
        for package in extra_packages:
            logger.info(f'- {package}')
    else:
        logger.info('Great! Your app is not having any extra package.')

    mismatch_packages = [
        package
        for package in req_packages
        if package in added_packages.keys()
        and added_packages[package] != req_packages[package]['version']
    ]
    if mismatch_packages:
        logger.warning(
            'Following packages are having mismatch between required and added version:'
        )
        for package in mismatch_packages:
            logger.info(
                f'- {package} requires version \'{req_packages[package]}\' but \'{added_packages[package]}\' is added'
            )
    else:
        logger.info('Great! Your app is not having any version-mismatching package.')



from nawah.config import Config
from nawah.classes import UnexpectedGatewayException, InvalidGatewayException

import logging, traceback

logger = logging.getLogger('nawah')


class Gateway:
    @staticmethod
    def send(*, gateway: str, **kwargs):
        if Config.test:
            logger.debug('Skipping \'Gateway.send\' action due to test mode.')
            return

        if gateway not in Config.gateways.keys():
            raise InvalidGatewayException(gateway=gateway)

        try:
            Config.gateways[gateway](**kwargs)
        except Exception as e:
            logger.error('Gateway call with following \'kwargs\' failed:')
            logger.error(kwargs)
            logger.error(traceback.format_exc())
            raise UnexpectedGatewayException(gateway=gateway)

        return



from nawah.base_module import BaseModule
from nawah.classes import ATTR, EXTN, METHOD, PERM, InvalidAttrException
from nawah.config import Config
from nawah.enums import Event
from nawah.utils import generate_dynamic_attr, validate_doc


async def attr_extn_val(
    *,
    mode,
    attr_name,
    attr_type,
    attr_val,
    skip_events,
    env,
    query,
    doc,
    scope,
):
    if type(doc['val']) == dict and '__extn' in doc['val'].keys():
        return {
            '__extn': EXTN(
                module=doc['val']['__extn']['__module'],
                attrs=doc['val']['__extn']['__attrs'],
                force=doc['val']['__extn']['__force'],
            ),
            '__val': doc['val']['__extn']['__val'],
        }

    raise InvalidAttrException(
        attr_name=attr_name, attr_type=attr_type, val_type=type(doc['val'])
    )


async def attr_query_mod_type(
    *,
    mode,
    attr_name,
    attr_type,
    attr_val,
    skip_events,
    env,
    query,
    doc,
    scope,
):
    if 'type' not in query or query['type'][0] == 'user_sys':
        raise InvalidAttrException(
            attr_name='type',
            attr_type=ATTR.LITERAL(literal=['global', 'user']),
            val_type=str,
        )


class Setting(BaseModule):
    '''`Setting` module module provides data type and controller for settings in Nawah eco-system. This is used by `User` module tp provide additional user-wise settings. It also allows for global-typed settings.'''

    collection = 'settings'
    attrs = {
        'user': ATTR.ID(desc='`_id` of `User` doc the doc belongs to.'),
        'var': ATTR.STR(
            desc='Name of the setting. This is unique for every `user` in the module.'
        ),
        'val': ATTR.ANY(desc='Value of the setting.'),
        'val_type': ATTR.DYNAMIC_ATTR(),
        'type': ATTR.LITERAL(
            desc='Type of the setting. This sets whether setting is global, or belong to user, and whether use can update it or not.',
            literal=['global', 'user', 'user_sys'],
        ),
    }
    diff = True
    unique_attrs = [('user', 'var', 'type')]
    extns = {
        'val': ATTR.TYPE(type=attr_extn_val),
    }
    methods = {
        'read': METHOD(
            permissions=[
                PERM(privilege='admin', query_mod={'$limit': 1}),
                PERM(
                    privilege='read',
                    query_mod={
                        'user': '$__user',
                        # 'type': ATTR.TYPE(type=attr_query_mod_type),
                        '$limit': 1,
                    },
                ),
            ],
            query_args=[
                {
                    '_id': ATTR.ID(),
                    'type': ATTR.LITERAL(literal=['global', 'user', 'user_sys']),
                },
                {
                    'var': ATTR.STR(),
                    'type': ATTR.LITERAL(literal=['global']),
                },
                {
                    'var': ATTR.STR(),
                    'user': ATTR.ID(),
                    'type': ATTR.LITERAL(literal=['user', 'user_sys']),
                },
            ],
        ),
        'create': METHOD(
            permissions=[
                PERM(privilege='admin'),
                PERM(privilege='create', doc_mod={'type': 'user'}),
            ]
        ),
        'update': METHOD(
            permissions=[
                PERM(privilege='admin', query_mod={'$limit': 1}),
                PERM(
                    privilege='update',
                    query_mod={'type': 'user', 'user': '$__user', '$limit': 1},
                    doc_mod={'var': None, 'val_type': None, 'type': None},
                ),
            ],
            query_args=[
                {
                    '_id': ATTR.ID(),
                    'type': ATTR.LITERAL(literal=['global', 'user', 'user_sys']),
                },
                {
                    'var': ATTR.STR(),
                    'type': ATTR.LITERAL(literal=['global']),
                },
                {
                    'var': ATTR.STR(),
                    'user': ATTR.ID(),
                    'type': ATTR.LITERAL(literal=['user', 'user_sys']),
                },
            ],
        ),
        'delete': METHOD(
            permissions=[PERM(privilege='admin', query_mod={'$limit': 1})],
            query_args=[{'_id': ATTR.ID()}, {'var': ATTR.STR()}],
        ),
        'retrieve_file': METHOD(
            permissions=[PERM(privilege='*', query_mod={'type': 'global'})],
            get_method=True,
        ),
    }

    async def on_create(self, results, skip_events, env, query, doc, payload):
        if doc['type'] in ['user', 'user_sys']:
            if (
                doc['user'] == env['session']['user']['_id']
                and doc['var'] in Config.user_doc_settings
            ):
                env['session']['user'][doc['var']] = doc['val']
        return (results, skip_events, env, query, doc, payload)

    async def pre_update(self, skip_events, env, query, doc, payload):
        for attr in doc.keys():
            if attr == 'val' or attr.startswith('val.'):
                val_attr = attr
                break
        else:
            raise self.exception(
                status=400,
                msg='Could not match doc with any of the required doc_args. Failed sets:[\'val\': Missing]',
                args={'code': 'INVALID_DOC'},
            )

        setting_results = await self.read(
            skip_events=[Event.PERM], env=env, query=query
        )
        if not setting_results['args']['count']:
            raise self.exception(
                status=400, msg='Invalid Setting doc', args={'code': 'INVALID_SETTING'}
            )

        setting = setting_results['args']['docs'][0]
        # [DOC] Attempt to validate val against Setting val_type
        try:
            exception_raised: Exception = None
            setting_val_type, _ = generate_dynamic_attr(
                dynamic_attr=setting['val_type']
            )
            await validate_doc(
                mode='update',
                doc=doc,
                attrs={'val': setting_val_type},
                skip_events=skip_events,
                env=env,
                query=query,
            )
        except Exception as e:
            exception_raised = e

        if exception_raised or doc[val_attr] == None:
            raise self.exception(
                status=400,
                msg=f'Invalid value for for Setting doc of type \'{type(doc[val_attr])}\' with required type \'{setting["val_type"]}\'',
                args={'code': 'INVALID_ATTR'},
            )

        return (skip_events, env, query, doc, payload)

    async def on_update(self, results, skip_events, env, query, doc, payload):
        # [TODO] Update according to the changes of Doc Opers
        try:
            if (
                query['type'][0] in ['user', 'user_sys']
                and query['user'][0] == env['session']['user']['_id']
                and query['var'][0] in Config.user_doc_settings
            ):
                if type(doc['val']) == dict and '$add' in doc['val'].keys():
                    env['session']['user'][query['var'][0]] += doc['val']['$add']
                elif type(doc['val']) == dict and '$multiply' in doc['val'].keys():
                    env['session']['user'][query['var'][0]] *= doc['val']['$multiply']
                elif type(doc['val']) == dict and '$append' in doc['val'].keys():
                    env['session']['user'][query['var'][0]].append(
                        doc['val']['$append']
                    )
                elif type(doc['val']) == dict and '$set_index' in doc['val'].keys():
                    env['session']['user'][query['var'][0]][doc['val']['$index']] = doc[
                        'val'
                    ]['$set_index']
                elif type(doc['val']) == dict and '$del_val' in doc['val'].keys():
                    for val in doc['val']['$del_val']:
                        env['session']['user'][query['var'][0]].remove(val)
                elif type(doc['val']) == dict and '$del_index' in doc['val'].keys():
                    del env['session']['user'][query['var'][0]][doc['val']['$index']]
                else:
                    env['session']['user'][query['var'][0]] = doc['val']
        except:
            pass
        return (results, skip_events, env, query, doc, payload)



from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Union

from nawah.enums import LOCALE_STRATEGY

if TYPE_CHECKING:
    from datetime import datetime

    from bson import ObjectId
    from motor.motor_asyncio import AsyncIOMotorClient
    from redis.Redis.json import JSON

    from nawah.base_module import BaseModule
    from nawah.classes import (ANALYTICS_EVENTS, ATTR, CLIENT_APP, JOB,
                               NAWAH_DOC, NAWAH_ENV, SYS_DOC, USER_SETTING)


class Config:
    debug: bool = False
    env: str
    port: int = 8081

    _sys_conn: 'AsyncIOMotorClient'
    _sys_cache: 'JSON' = None
    _sys_env: 'NAWAH_ENV'
    _sys_docs: Dict['ObjectId', 'SYS_DOC'] = {}
    _jobs_base: 'datetime'

    _nawah_version: str
    packages_api_levels: Dict[str, str] = {}
    packages_versions: Dict[str, str] = {}

    _app_name: str
    _app_version: str
    _app_default_package: str
    _app_path: str
    _app_packages: Dict[str, str]

    test: bool = False

    emulate_test: bool = False
    force_admin_check: bool = False

    generate_ref: bool = False
    _api_ref: str

    generate_models: bool = False
    _api_models: str

    vars_types: Dict[str, Union['ATTR', Dict[str, Any]]] = {}
    vars: Dict[str, Any] = {}

    client_apps: Dict[str, 'CLIENT_APP'] = {}

    analytics_events: 'ANALYTICS_EVENTS' = {
        'app_conn_verified': True,
        'session_conn_auth': True,
        'session_user_auth': True,
        'session_conn_reauth': True,
        'session_user_reauth': True,
        'session_conn_deauth': True,
        'session_user_deauth': True,
    }

    conn_timeout: int = 120
    quota_anon_min: int = 40
    quota_auth_min: int = 100
    quota_ip_min: int = 500
    file_upload_limit: int = -1
    file_upload_timeout: int = 300

    data_server: str = 'mongodb://localhost'
    data_name: str = 'nawah_data'
    data_ssl: bool = False
    data_ca_name: Optional[str] = None
    data_ca: Optional[str] = None
    data_disk_use: bool = False

    data_azure_mongo: bool = False

    cache_server: Optional[str] = None
    cache_db: Optional[int] = 0

    locales: List[str] = ['ar_AE', 'en_AE']
    locale: str = 'ar_AE'
    locale_strategy: 'LOCALE_STRATEGY' = LOCALE_STRATEGY.DUPLICATE

    admin_doc: 'NAWAH_DOC' = {}
    admin_password: str = '__0xADMIN'

    anon_token: str = '__ANON_TOKEN_f00000000000000000000012'
    anon_privileges: Dict[str, List[str]] = {}

    user_attrs: Dict[str, 'ATTR'] = {}
    user_settings: Dict[str, 'USER_SETTING'] = {}
    user_doc_settings: List[str] = []

    groups: List[Dict[str, Any]] = []
    default_privileges: Dict[str, List[str]] = {}

    data_indexes: List[Dict[str, Any]] = []

    docs: List['SYS_DOC'] = []

    l10n: Dict[str, Dict[str, Any]] = {}

    jobs: Dict[str, 'JOB'] = {}

    gateways: Dict[str, Callable] = {}

    types: Dict[str, Callable] = {}

    modules: Dict[str, 'BaseModule'] = {}
    modules_packages: Dict[str, List[str]] = {}



import datetime
from asyncio import Task
from typing import (TYPE_CHECKING, Any, AsyncGenerator, Callable, Coroutine,
                    Dict, List, Literal, Optional, TypedDict, Union)

from aiohttp.web import WebSocketResponse
from motor.motor_asyncio import AsyncIOMotorClient

from nawah.enums import Event

NAWAH_EVENTS = List[Event]

WATCH_TASK = TypedDict(
    'WATCH_TASK',
    {
        'watch': Coroutine[Any, Any, None],
        'task': Task,
        'stream': Any,
    },
    total=False,
)

NAWAH_ENV_QUOTA = TypedDict(
    'NAWAH_ENV_QUOTA',
    {
        'counter': int,
        'last_check': datetime.datetime,
    },
)

NAWAH_ENV = TypedDict(
    'NAWAH_ENV',
    {
        'id': int,
        'init': bool,
        'conn': AsyncIOMotorClient,
        'REMOTE_ADDR': str,
        'HTTP_USER_AGENT': str,
        'HTTP_ORIGIN': str,
        'client_app': str,
        'session': Dict,
        'last_call': datetime.datetime,
        'ws': WebSocketResponse,
        'watch_tasks': Dict[str, WATCH_TASK],
        'quota': NAWAH_ENV_QUOTA,
    },
    total=False,
)

NAWAH_QUERY_SPECIAL_GROUP = TypedDict(
    'NAWAH_QUERY_SPECIAL_GROUP', {'by': str, 'count': int}
)

NAWAH_QUERY_SPECIAL_GEO_NEAR = TypedDict(
    'NAWAH_QUERY_SPECIAL_GEO_NEAR', {'val': str, 'attr': str, 'dist': int}
)

NAWAH_QUERY_SPECIAL = TypedDict(
    'NAWAH_QUERY_SPECIAL',
    {
        '$search': Optional[str],
        '$sort': Optional[Dict[str, Literal[1, -1]]],
        '$skip': Optional[int],
        '$limit': Optional[int],
        '$extn': Optional[Union[Literal[False], List[str]]],
        '$attrs': Optional[List[str]],
        '$group': Optional[List[NAWAH_QUERY_SPECIAL_GEO_NEAR]],
        '$geo_near': Optional[NAWAH_QUERY_SPECIAL_GEO_NEAR],
    },
    total=False,
)

NAWAH_QUERY = List[  # type: ignore
    Union[
        'NAWAH_QUERY',  # type: ignore
        Union[
            Dict[
                str,
                Union[
                    'NAWAH_QUERY',  # type: ignore
                    Any,
                    Union[
                        Dict[Literal['$eq'], Any],
                        Dict[Literal['$ne'], Any],
                        Dict[Literal['$gt'], Union[int, str]],
                        Dict[Literal['$gte'], Union[int, str]],
                        Dict[Literal['$lt'], Union[int, str]],
                        Dict[Literal['$lte'], Union[int, str]],
                        Dict[Literal['$bet'], Union[List[int], List[str]]],
                        Dict[Literal['$all'], List[Any]],
                        Dict[Literal['$in'], List[Any]],
                        Dict[Literal['$nin'], List[Any]],
                        Dict[Literal['$regex'], str],
                    ],
                ],
            ],
            NAWAH_QUERY_SPECIAL,
        ],
    ]
]

NAWAH_DOC = Dict[
    str,
    Union[
        Dict[
            str,
            Union[
                Dict[Literal['$add'], int],
                Dict[Literal['$multiply'], int],
                Dict[Literal['$append'], Any],
                Dict[Literal['$set_index'], Dict[int, Any]],
                Dict[Literal['$set_index'], Dict[int, Any]],
                Dict[Literal['$del_val'], Any],
                Dict[Literal['$del_index'], int],
                Any,
            ],
        ],
        Any,
    ],
]

IP_QUOTA = TypedDict(
    'IP_QUOTA',
    {
        'counter': int,
        'last_check': datetime.datetime,
    },
)



from nawah.config import Config
from nawah.utils import _extract_attr
from nawah.classes import (
    SYS_DOC,
    JOB,
    InvalidModuleException,
    InvalidLocaleException,
    InvalidLocaleTermException,
    InvalidVarException,
)

from typing import Dict, List, Any, TYPE_CHECKING

import logging

if TYPE_CHECKING:
    from nawah.base_module import BaseModule

logger = logging.getLogger('nawah')


class Registry:
    docs: List[SYS_DOC] = Config.docs
    jobs: Dict[str, JOB] = Config.jobs

    @staticmethod
    def module(module: str) -> 'BaseModule':
        try:
            return Config.modules[module]
        except KeyError:
            raise InvalidModuleException(module=module)

    @staticmethod
    def l10n(*, locale: str, term: str):
        if locale not in Config.l10n.keys():
            raise InvalidLocaleException(locale=locale)

        try:
            if '.' in term:
                return _extract_attr(scope=Config.l10n[locale], attr_path=term)
            else:
                return Config.l10n[locale][term]
        except (KeyError, TypeError):
            raise InvalidLocaleTermException(locale=locale, term=term)

    @staticmethod
    def var(var: str):
        try:
            if '.' in var:
                return _extract_attr(scope=Config.vars, attr_path=var)
            else:
                return Config.vars[var]
        except (KeyError, TypeError):
            raise InvalidVarException(var=var)



'''Provides Core Nawah module'''

from nawah.base_module import BaseModule
from nawah.classes import ATTR, METHOD, PERM
from nawah.config import Config


class Core(BaseModule):
    '''`Core` module provides access to ADMIN user to fetch `Nawah` instance.'''

    methods = {
        'retrieve_config': METHOD(
            permissions=[PERM(privilege='admin')],
            query_args={'config_attr': ATTR.STR()},
        ),
    }

    async def retrieve_config(self, query):
        '''Returns current value of provided `Config Attr`'''

        return self.status(
            status=200,
            msg='Config Attr value retrieved.',
            args={'value': getattr(Config, query['config_attr'][0])},
        )



from nawah.base_module import BaseModule
from nawah.enums import Event
from nawah.classes import ATTR, PERM, METHOD
from nawah.config import Config
from nawah.registry import Registry
from nawah.utils import _extract_attr, validate_attr

from bson import ObjectId

import base64


class File(BaseModule):
    '''`File` module provides functionality for `File Upload Workflow`.'''

    collection = 'files'
    attrs = {
        'user': ATTR.ID(desc='`_id` of `User` doc file belongs to.'),
        'file': ATTR.FILE(desc='File object.'),
        'create_time': ATTR.DATETIME(
            desc='Python `datetime` ISO format of the doc creation.'
        ),
    }
    methods = {
        'read': METHOD(permissions=[PERM(privilege='__sys')]),
        'create': METHOD(
            permissions=[PERM(privilege='create')],
            post_method=True,
        ),
        'delete': METHOD(permissions=[PERM(privilege='__sys')]),
    }

    async def on_read(self, results, skip_events, env, query, doc, payload):
        for i in range(len(results['docs'])):
            results['docs'][i]['file']['lastModified'] = int(
                results['docs'][i]['file']['lastModified']
            )
        return (results, skip_events, env, query, doc, payload)

    async def pre_create(self, skip_events, env, query, doc, payload):
        if (
            Config.file_upload_limit != -1
            and len(doc['file']) > Config.file_upload_limit
        ):
            raise self.exception(
                status=400,
                msg=f'File size is beyond allowed limit.',
                args={
                    'code': 'INVALID_SIZE',
                    'attr': doc['__attr'].decode('utf-8'),
                    'name': doc['name'].decode('utf-8'),
                },
            )
        if (module := doc['__module'].decode('utf-8')) not in Config.modules.keys():
            raise self.exception(
                status=400,
                msg=f'Invalid module \'{module}\'',
                args={'code': 'INVALID_MODULE'},
            )

        try:
            attr_type = _extract_attr(
                scope=Registry.module(module).attrs,
                attr_path='$__' + (attr := doc['__attr'].decode('utf-8')),
            )
            doc = {
                'file': {
                    'name': doc['name'].decode('utf-8'),
                    'type': doc['type'].decode('utf-8'),
                    'size': len(doc['file']),
                    'lastModified': int(doc['lastModified'].decode('utf-8')),
                    'content': doc['file'],
                },
            }
            try:
                attr_val = doc['file']
                if attr_type._type == 'LIST':
                    attr_val = [doc['file']]
                await validate_attr(
                    mode='create',
                    attr_name=attr,
                    attr_type=attr_type,
                    attr_val=attr_val,
                )
            except:
                raise self.exception(
                    status=400,
                    msg=f'Invalid file for \'{attr}\' of module \'{module}\'',
                    args={'code': 'INVALID_FILE'},
                )

        except:
            raise self.exception(
                status=400,
                msg=f'Invalid attr \'{attr}\' of module \'{module}\'',
                args={'code': 'INVALID_ATTR'},
            )

        return (skip_events, env, query, doc, payload)



from nawah import __version__

from typing import Dict, Literal, Any, Optional

import argparse, os, logging, datetime, sys, subprocess, asyncio, traceback, shutil, urllib.request, re, tarfile, string, random, tempfile, pkgutil, glob


def nawah_cli():
    from ._logging import logger
    from ._create import create
    from ._launch import launch
    from ._packages import packages_audit, packages_install, _packages_add, _packages_rm
    from ._generate import generate_ref, generate_models

    if sys.version_info.major != 3 or sys.version_info.minor < 8:
        print('Nawah framework CLI can only run with Python >= 3.8. Exiting.')
        exit(1)

    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--version',
        help='Show Nawah framework version and exit',
        action='version',
        version=f'Nawah framework v{__version__}',
    )

    subparsers = parser.add_subparsers(
        title='Command',
        description='Nawah framework CLI command to run',
        dest='command',
    )

    parser_create = subparsers.add_parser('create', help='Create new Nawah app')
    parser_create.set_defaults(func=create)
    parser_create.add_argument('app_name', type=str, help='Name of the app to create')
    parser_create.add_argument(
        'app_path',
        type=str,
        nargs='?',
        help='Path to create new Nawah app. [default .]',
        default='.',
    )
    parser_create.add_argument(
        '--default-config',
        help='Create new Nawah app with default config',
        action='store_true',
    )

    parser_launch = subparsers.add_parser('launch', help='Launch Nawah app')
    parser_launch.set_defaults(func=launch)
    parser_launch.add_argument('--env', help='Choose specific env')
    parser_launch.add_argument('--debug', help='Enable debug mode', action='store_true')
    parser_launch.add_argument(
        '--log',
        help='Enable debug mode and log all debug messages to log file',
        action='store_true',
    )
    parser_launch.add_argument('-p', '--port', help='Set custom port [default 8081]')
    parser_launch.add_argument(
        '--force-admin-check',
        help='Force ADMIN doc checked and updated, if ADMIN doc is changed',
        action='store_true',
    )

    parser_packages = subparsers.add_parser(
        'packages', help='Manage Nawah app packages'
    )
    parser_packages.set_defaults(func=lambda _: None)
    packages_subparser = parser_packages.add_subparsers(
        title='Packages Command',
        description='Packages command to run',
        dest='packages_command',
    )

    parser_packages_install = packages_subparser.add_parser(
        'install', help='Install Nawah app packages'
    )
    parser_packages_install.set_defaults(func=packages_install)

    parser_packages_add = packages_subparser.add_parser(
        'add', help='Add package to Nawah app'
    )
    parser_packages_add.set_defaults(
        func=lambda args: _packages_add(
            package_name=args.package_name,
            source=args.source,
            version=args.version,
            auth=args.auth,
        )
    )
    parser_packages_add.add_argument(
        'package_name', help='Package name to add to Nawah app'
    )
    parser_packages_add.add_argument(
        '--source',
        help='Package source [default https://gitlab.com/api/v4/projects/24381550/packages/pypi/simple]',
        default='https://gitlab.com/api/v4/projects/24381550/packages/pypi/simple',
    )
    parser_packages_add.add_argument(
        '--version',
        help='Package version (repo tag name) to install [default latest]',
        default='latest',
    )
    parser_packages_add.add_argument(
        '--auth',
        help='String representing colon-separated username and password combination to authorise the source',
    )

    parser_packages_rm = packages_subparser.add_parser(
        'rm', help='Remove package from Nawah app'
    )
    parser_packages_rm.set_defaults(
        func=lambda args: _packages_rm(
            package_name=args.package_name, confirm=not args.y
        )
    )
    parser_packages_rm.add_argument(
        'package_name', help='Package name to remove from Nawah app'
    )
    parser_packages_rm.add_argument('-y', help='Skip confirmation', action='store_true')

    parser_packages_audit = packages_subparser.add_parser(
        'audit', help='Audit packages in Nawah app'
    )
    parser_packages_audit.set_defaults(func=packages_audit)

    parser_ref = subparsers.add_parser(
        'generate_ref', help='Generate Nawah app reference'
    )
    parser_ref.set_defaults(func=generate_ref)
    parser_ref.add_argument('--debug', help='Enable debug mode', action='store_true')

    parser_ref = subparsers.add_parser(
        'generate_models', help='Generate Nawah app models'
    )
    parser_ref.set_defaults(func=generate_models)
    parser_ref.add_argument('format', help='Format of models', choices=['js', 'ts'])
    parser_ref.add_argument('--debug', help='Enable debug mode', action='store_true')

    args = parser.parse_args()

    if args.command:
        if args.command == 'packages' and not args.packages_command:
            parser_packages.print_help()
        else:
            args.func(args)
    else:
        parser.print_help()



'''Provides `reset_cache` Utility'''

from nawah.config import Config


def reset_cache(channel: str) -> None:
    '''Resets specific cache `channel` by deleting it from active Redis db'''

    Config._sys_cache.delete(channel, '.')



from nawah import __version__

from typing import Literal

import argparse, os, sys, logging, asyncio, datetime, traceback

from ._logging import logger, handler, formatter


def launch(
    args: argparse.Namespace,
    custom_launch: Literal['test', 'generate_ref', 'generate_models'] = None,
):

    # [DOC] Update Config with Nawah CLI args
    from nawah.config import Config
    from nawah.utils import _process_config

    Config._nawah_version = __version__
    if custom_launch not in ['generate_ref', 'generate_models']:
        Config.env = args.env
    if not custom_launch:
        Config.force_admin_check = args.force_admin_check

    # [DOC] Check for debug CLI Arg
    if args.debug:
        Config.debug = True
        logger.setLevel(logging.DEBUG)
    # [DOC] Check for log CLI Arg
    if not custom_launch and args.log:
        logger.removeHandler(handler)
        if not os.path.exists(os.path.join('.', 'logs')):
            os.makedirs(os.path.join('.', 'logs'))
        file_handler = logging.FileHandler(
            filename=os.path.join(
                '.',
                'logs',
                f'{datetime.datetime.utcnow().strftime("%d-%b-%Y")}.log',
            )
        )
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)
        logger.setLevel(logging.DEBUG)

    from nawah.app import run_app
    import json

    try:
        try:
            sys.path.append('.')
            nawah_app = __import__('nawah_app')
            app_config = nawah_app.config

            with open('packages.json') as f:
                packages = json.loads(f.read())

            Config._app_packages = {k: v['version'] for k, v in packages.items()}
        except ModuleNotFoundError:
            logger.error(f'No \'nawah_app.py\' file found in CWD. Exiting.')
            exit(1)
        except AttributeError:
            logger.error(
                f'File \'nawah_app.py\' was found but it doesn\'t have \'config\' method. Exiting.'
            )
            exit(1)
        logger.info(
            f'Found app \'{app_config.name} (v{app_config.version})\'. Attempting to load App Config.'
        )
        Config._app_name = app_config.name
        Config._app_version = app_config.version
        Config._app_default_package = app_config.default_package

        # [DOC] Check all required values of App Config are present
        if (
            not Config._app_name
            or not Config._app_version
            or not Config._app_default_package
        ):
            logger.error(
                f'App Config are missing at least one of the following \'name, version, default_package\'. Exiting.'
            )
            exit(1)

        Config._app_path = os.path.realpath('.')
        # [DOC] Read app_config and update Config accordingly
        # [DOC] Check envs, env
        if custom_launch not in ['generate_ref', 'generate_models'] and app_config.envs:
            if not args.env and not app_config.env:
                logger.error(
                    'App Config Attr \'envs\' found, but no \'env\' App Config Attr, or CLI Attr were defined.'
                )
                exit(1)
            if args.env:
                if args.env in app_config.envs.keys():
                    logger.info(
                        f'Setting \'env\' Config Attr to \'env\' CLI Arg value \'{args.env}\''
                    )
                else:
                    logger.error(
                        f'Found value \'{args.env}\' for \'env\' CLI Arg, but not defined in \'envs\' App Config Attr. Exiting.'
                    )
                    exit(1)
            else:
                if app_config.env in app_config.envs.keys():
                    logger.info(
                        f'Setting \'env\' Config Attr to \'env\' App Config Attr value \'{app_config.env}\''
                    )
                    Config.env = app_config.env
                elif app_config.env.startswith('$__env.'):
                    logger.info(
                        'Found Env Variable for \'env\' App Config Attr. Attempting to process it.'
                    )
                    env_env_var = app_config.env.replace('$__env.', '')
                    env = os.getenv(env_env_var)
                    if env:
                        logger.info(
                            f'Setting \'env\' Config Attr to Env Variable \'{env_env_var}\' value \'{env}\'.'
                        )
                        Config.env = env
                    else:
                        logger.error(
                            f'No value found for Env Variable \'{env_env_var}\'. Exiting.'
                        )
                        exit(1)
                else:
                    logger.error(
                        f'Found value \'{args.env}\' for \'env\' CLI Arg, but not defined in \'envs\' App Config Attr. Exiting.'
                    )
                    exit(1)
            logger.info(
                f'Beginning to extract Config Attrs defined in selected \'env\', \'{Config.env}\', to App Config Attrs.'
            )
            for config_attr in dir(app_config.envs[Config.env]):
                if (
                    config_attr.startswith('__')
                    or getattr(app_config.envs[Config.env], config_attr) == None
                ):
                    continue
                logger.info(
                    f'Extracting \'{config_attr}\' Config Attr to App Config Attr'
                )
                setattr(
                    app_config,
                    config_attr,
                    getattr(app_config.envs[Config.env], config_attr),
                )
                setattr(
                    Config,
                    config_attr,
                    getattr(app_config.envs[Config.env], config_attr),
                )
        # [DOC] Check port Config Attr
        if not custom_launch and app_config.port:
            if args.port:
                logger.info(
                    f'Ignoring \'port\' App Config Attr in favour of \'port\' CLI Arg with value \'{args.port}\'.'
                )
                try:
                    Config.port = int(args.port)
                except:
                    logger.error(f'Port should be in integer format. Exiting.')
                    exit(1)
            else:
                logger.info('Found \'port\' App Config Attr. Attempting to process it.')
                if type(app_config.port) == int:
                    Config.port = app_config.port
                    logger.info(f'Setting \'port\' Config Attr to \'{Config.port}\'.')
                elif type(app_config.port) == str and app_config.port.startswith(
                    '$__env.'
                ):
                    logger.info(
                        'Found Env Variable for \'port\' App Config Attr. Attempting to process it.'
                    )
                    port_env_var = app_config.port.replace('$__env.', '')
                    port = os.getenv(port_env_var)
                    if port:
                        logger.info(
                            f'Setting \'port\' Config Attr to Env Variable \'{port_env_var}\' value \'{port}\'.'
                        )
                        try:
                            Config.port = int(port)
                        except:
                            logger.error(
                                f'Env Variable \'{port_env_var}\' value \'{port}\' can\'t be converted to integer. Exiting.'
                            )
                            exit(1)
                    else:
                        logger.error(
                            f'No value found for Env Variable \'{port_env_var}\'. Exiting.'
                        )
                        exit(1)
                else:
                    logger.error(
                        f'Invalid value type for \'port\' Config Attr with value \'{app_config.port}\'. Exiting.'
                    )
                    exit(1)
        # [DOC] Check debug Config Attr
        if app_config.debug:
            if args.debug:
                logger.info(
                    f'Ignoring \'debug\' App Config Attr in favour of \'debug\' CLI Arg with value \'{args.debug}\'.'
                )
                Config.debug = args.debug
            else:
                logger.info(
                    'Found \'debug\' App Config Attr. Attempting to process it.'
                )
                if type(app_config.debug) == bool:
                    Config.debug = app_config.debug
                    logger.info(f'Setting \'debug\' Config Attr to \'{Config.debug}\'.')
                elif type(app_config.debug) == str and app_config.debug.startswith(
                    '$__env.'
                ):
                    logger.info(
                        'Found Env Variable for \'debug\' App Config Attr. Attempting to process it.'
                    )
                    debug_env_var = app_config.debug.replace('$__env.', '')
                    debug = os.getenv(debug_env_var)
                    if debug:
                        logger.info(
                            f'Setting \'debug\' Config Attr to Env Variable \'{debug_env_var}\' as \'True\'.'
                        )
                        Config.debug = True
                    else:
                        logger.info(
                            f'No value found for Env Variable \'{debug_env_var}\'. Setting \'debug\' to \'False\'.'
                        )
                        Config.debug = False
                else:
                    logger.error(
                        f'Invalid value type for \'debug\' Config Attr with value \'{app_config.debug}\'. Exiting.'
                    )
                    exit(1)
                if Config.debug:
                    logger.setLevel(logging.DEBUG)
        # [DOC] Check force_admin_check Config Attr
        if not custom_launch and app_config.force_admin_check:
            if args.force_admin_check:
                logger.info(
                    f'Ignoring \'force_admin_check\' App Config Attr in favour of \'force_admin_check\' CLI Arg with value \'{args.force_admin_check}\'.'
                )
                Config.force_admin_check = True
            else:
                logger.info(
                    'Found \'force_admin_check\' App Config Attr. Attempting to process it.'
                )
                if type(app_config.force_admin_check) == bool:
                    Config.force_admin_check = app_config.force_admin_check
                    logger.info(
                        f'Setting \'force_admin_check\' Config Attr to \'{Config.force_admin_check}\'.'
                    )
                elif type(
                    app_config.force_admin_check
                ) == str and app_config.force_admin_check.startswith('$__env.'):
                    logger.info(
                        'Found Env Variable for \'force_admin_check\' App Config Attr. Attempting to process it.'
                    )
                    check_env_var = app_config.force_admin_check.replace('$__env.', '')
                    check = os.getenv(check_env_var)
                    if check:
                        logger.info(
                            f'Setting \'force_admin_check\' Config Attr to Env Variable \'{check_env_var}\' as \'True\'.'
                        )
                        Config.force_admin_check = True
                    else:
                        logger.info(
                            f'No value found for Env Variable \'{check_env_var}\'. Setting \'force_admin_check\' to \'False\'.'
                        )
                        Config.force_admin_check = False
                else:
                    logger.error(
                        f'Invalid value type for \'force_admin_check\' Config Attr with value \'{app_config.force_admin_check}\'. Exiting.'
                    )
                    exit(1)

        # [DOC] Process other app config attrs as PACKAGE_CONFIG
        _process_config(config=app_config)
    except SystemExit as e:
        exit(e)
    except:
        logger.error(
            'An unexpected exception happened while attempting to process Nawah app. Exception details:'
        )
        logger.error(traceback.format_exc())
        logger.error('Exiting.')
        exit(1)

    if not custom_launch:
        asyncio.run(run_app())
    elif custom_launch == 'generate_ref':

        async def _():
            from nawah.utils import _import_modules, _generate_ref

            await _import_modules()
            _generate_ref()

        asyncio.run(_())
    elif custom_launch == 'generate_models':

        async def _():
            from nawah.utils import _import_modules, _generate_models

            await _import_modules()
            _generate_models()

        asyncio.run(_())



from nawah.config import Config
from nawah.classes import ATTR, EXTN

from typing import Dict, List

import logging, datetime, re, inspect

logger = logging.getLogger('nawah')


def _generate_ref():
    # [DOC] Initialise _api_ref Config Attr
    Config._api_ref = '# API Reference\n'
    # [DOC] Iterate over packages in ascending order
    for package in sorted(Config.modules_packages.keys()):
        # [DOC] Add package header
        Config._api_ref += f'- - -\n## Package: {package.replace("modules.", "")}\n'
        if not len(Config.modules_packages[package]):
            Config._api_ref += f'No modules\n'
        # [DOC] Iterate over package modules in ascending order
        for module in sorted(Config.modules_packages[package]):
            # [DOC] Add module header
            Config._api_ref += f'### Module: {module}\n'
            # [DOC] Add module description
            Config._api_ref += f'{Config.modules[module].__doc__}\n'
            # [DOC] Add module attrs header
            Config._api_ref += '#### Attrs\n'
            # [DOC] Iterate over module attrs to add attrs types, defaults (if any)
            for attr in Config.modules[module].attrs.keys():
                attr_ref = f'* {attr}:\n'
                if Config.modules[module].attrs[attr]._desc:
                    attr_ref += f'  * {Config.modules[module].attrs[attr]._desc}\n'
                attr_ref += f'  * Type: `{Config.modules[module].attrs[attr]}`\n'
                for default_attr in Config.modules[module].defaults.keys():
                    if (
                        default_attr == attr
                        or default_attr.startswith(f'{attr}.')
                        or default_attr.startswith(f'{attr}:')
                    ):
                        if type(Config.modules[module].defaults[default_attr]) == ATTR:
                            attr_ref += f'  * Default [{default_attr}]:\n'
                            attr_ref += f'	* Attr Type TYPE: `{Config.modules[module].defaults[default_attr]._args["type"]}`\n'
                        else:
                            attr_ref += f'  * Default [{default_attr}]: {Config.modules[module].defaults[default_attr]}\n'
                Config._api_ref += attr_ref
            if Config.modules[module].diff:
                Config._api_ref += f'#### Attrs Diff: {Config.modules[module].diff}\n'
            # [DOC] Add module methods
            Config._api_ref += '#### Methods\n'
            for method in Config.modules[module].methods.keys():
                Config._api_ref += f'##### Method: {method}\n'
                Config._api_ref += f'* Permissions Sets:\n'
                for permission in Config.modules[module].methods[method].permissions:
                    Config._api_ref += f'  * {permission.privilege}\n'
                    # [DOC] Add Query Modifier
                    if permission.query_mod:
                        Config._api_ref += f'	* Query Modifier:\n'
                        if type(permission.query_mod) == dict:
                            permission.query_mod = [permission.query_mod]
                        for i in range(len(permission.query_mod)):
                            Config._api_ref += f'	  * Set {i}:\n'
                            # [TODO] Improve nested list sets
                            if type(permission.query_mod[i]) != dict:
                                Config._api_ref += (
                                    f'		* List: {permission.query_mod[i]}\n'
                                )
                                continue
                            for attr in permission.query_mod[i].keys():
                                if type(permission.query_mod[i][attr]) == ATTR:
                                    Config._api_ref += f'		* {attr}:\n'
                                    Config._api_ref += f'		  * Attr Type TYPE: {permission.query_mod[i][attr]._args["type"]}\n'
                                else:
                                    Config._api_ref += (
                                        f'		* {attr}: {permission.query_mod[i][attr]}\n'
                                    )
                    else:
                        Config._api_ref += f'	* Query Modifier: None\n'
                    # [DOC] Add Doc Modifier
                    if permission.doc_mod:
                        Config._api_ref += f'	* Doc Modifier:\n'
                        if type(permission.doc_mod) == dict:
                            permission.doc_mod = [permission.doc_mod]
                        for i in range(len(permission.doc_mod)):
                            Config._api_ref += f'	  * Set {i}:\n'
                            for attr in permission.doc_mod[i].keys():
                                if type(permission.doc_mod[i][attr]) == ATTR:
                                    Config._api_ref += f'		* {attr}:\n'
                                    Config._api_ref += f'		  * Attr Type TYPE: `{permission.doc_mod[i][attr]._args["type"]}`\n'
                                else:
                                    Config._api_ref += (
                                        f'		* {attr}: {permission.doc_mod[i][attr]}\n'
                                    )
                    else:
                        Config._api_ref += f'	* Doc Modifier: None\n'
                # [DOC] Add Query Args
                if Config.modules[module].methods[method].query_args:
                    Config._api_ref += f'* Query Args Sets:\n'
                    for query_args_set in (
                        Config.modules[module].methods[method].query_args
                    ):
                        Config._api_ref += f'  * `{query_args_set}`\n'
                else:
                    Config._api_ref += f'* Query Args Sets: None\n'
                # [DOC] Add Doc Args
                if Config.modules[module].methods[method].doc_args:
                    Config._api_ref += f'* DOC Args Sets:\n'
                    for doc_args_set in Config.modules[module].methods[method].doc_args:
                        Config._api_ref += f'  * `{doc_args_set}`\n'
                else:
                    Config._api_ref += f'* Doc Args Sets: None\n'
            # [DOC] Add module extns
            if Config.modules[module].extns.keys():
                Config._api_ref += '#### Extended Attrs\n'
                for attr in Config.modules[module].extns.keys():
                    Config._api_ref += f'* {attr}:\n'
                    if type(Config.modules[module].extns[attr]) == EXTN:
                        Config._api_ref += f'  * Module: \'{Config.modules[module].extns[attr].module}\'\n'
                        Config._api_ref += f'  * Extend Attrs: \'{Config.modules[module].extns[attr].attrs}\'\n'
                        Config._api_ref += f'  * Force: \'{Config.modules[module].extns[attr].force}\'\n'
                    elif type(Config.modules[module].extns[attr]) == ATTR:
                        Config._api_ref += f'  * Attr Type TYPE: `{Config.modules[module].extns[attr]._args["type"]}`\n'
            else:
                Config._api_ref += '#### Extended Attrs: None\n'
            # [DOC] Add module cache sets
            if Config.modules[module].cache:
                Config._api_ref += '#### Cache Sets\n'
                for i in range(len(Config.modules[module].cache)):
                    Config._api_ref += f'* Set {i}:\n'
                    Config._api_ref += f'  * CACHE condition: `{_extract_lambda_body(Config.modules[module].cache[i].condition)}`\n'
                    Config._api_ref += (
                        f'  * CACHE period: {Config.modules[module].cache[i].period}\n'
                    )
            else:
                Config._api_ref += '#### Cache Sets: None\n'
            # [DOC] Add module analytics sets
            if Config.modules[module].analytics:
                Config._api_ref += '#### Analytics Sets\n'
                for i in range(len(Config.modules[module].analytics)):
                    Config._api_ref += f'* Set {i}:\n'
                    Config._api_ref += f'  * ANALYTIC condition: `{_extract_lambda_body(Config.modules[module].analytics[i].condition)}`\n'
                    Config._api_ref += f'  * ANALYTIC doc: `{_extract_lambda_body(Config.modules[module].analytics[i].doc)}`\n'
            else:
                Config._api_ref += '#### Analytics Sets: None\n'
    import os

    ref_file = os.path.join(
        Config._app_path,
        'refs',
        f'NAWAH_API_REF_{datetime.datetime.utcnow().strftime("%d-%b-%Y")}.md',
    )
    with open(ref_file, 'w') as f:
        f.write(Config._api_ref)
        logger.info(f'API reference generated and saved to: \'{ref_file}\'. Exiting.')
        exit(0)


def _extract_lambda_body(lambda_func):
    lambda_body = re.sub(
        r'^[a-z]+\s*=\s*lambda\s', '', inspect.getsource(lambda_func).strip()
    )
    if lambda_body.endswith(','):
        lambda_body = lambda_body[:-1]
    return lambda_body



from nawah.config import Config
from nawah.classes import ATTR

from bson import ObjectId
from typing import Any

import logging, random, datetime, re, math

logger = logging.getLogger('nawah')


def generate_attr(*, attr_type: ATTR) -> Any:
    attr_val: Any

    if attr_type._type == 'ANY':
        return '__any'

    elif attr_type._type == 'ACCESS':
        return {'anon': True, 'users': [], 'groups': []}

    elif attr_type._type == 'BOOL':
        attr_val = random.choice([True, False])
        return attr_val

    elif attr_type._type == 'COUNTER':
        counter_groups = re.findall(
            r'(\$__(?:values:[0-9]+|counters\.[a-z0-9_]+))', attr_type._args['pattern']
        )
        attr_val = attr_type._args['pattern']
        for group in counter_groups:
            for group in counter_groups:
                if group.startswith('$__values:'):
                    value_callable = attr_type._args['values'][
                        int(group.replace('$__values:', ''))
                    ]
                    attr_val = attr_val.replace(
                        group,
                        str(value_callable(skip_events=[], env={}, query=[], doc={})),
                    )
                elif group.startswith('$__counters.'):
                    attr_val = attr_val.replace(group, str(42))
        return attr_val

    elif attr_type._type == 'DATE':
        if attr_type._args['ranges']:
            datetime_range = attr_type._args['ranges'][0]
            # [DOC] Be lazy! find a whether start, end of range is a datetime and base the value on it
            if datetime_range[0][0] in ['+', '-'] and datetime_range[1][0] in [
                '+',
                '-',
            ]:
                # [DOC] Both start, end are dynamic, process start
                datetime_range_delta = {}
                if datetime_range[0][-1] == 'd':
                    datetime_range_delta = {'days': int(datetime_range[0][:-1])}
                elif datetime_range[0][-1] == 'w':
                    datetime_range_delta = {'weeks': int(datetime_range[0][:-1])}
                attr_val = (
                    (
                        datetime.datetime.utcnow()
                        + datetime.timedelta(**datetime_range_delta)
                    )
                    .isoformat()
                    .split('T')[0]
                )
            else:
                if datetime_range[0][0] not in ['+', '-']:
                    attr_val = datetime_range[0]
                else:
                    attr_val = (
                        (
                            datetime.datetime.fromisoformat(datetime_range[1])
                            - datetime.timedelta(days=1)
                        )
                        .isoformat()
                        .split('T')[0]
                    )
        else:
            attr_val = datetime.datetime.utcnow().isoformat().split('T')[0]
        return attr_val

    elif attr_type._type == 'DATETIME':
        if attr_type._args['ranges']:
            datetime_range = attr_type._args['ranges'][0]
            # [DOC] Be lazy! find a whether start, end of range is a datetime and base the value on it
            if datetime_range[0][0] in ['+', '-'] and datetime_range[1][0] in [
                '+',
                '-',
            ]:
                # [DOC] Both start, end are dynamic, process start
                datetime_range_delta = {}
                if datetime_range[0][-1] == 'd':
                    datetime_range_delta = {'days': int(datetime_range[0][:-1])}
                elif datetime_range[0][-1] == 's':
                    datetime_range_delta = {'seconds': int(datetime_range[0][:-1])}
                elif datetime_range[0][-1] == 'm':
                    datetime_range_delta = {'minutes': int(datetime_range[0][:-1])}
                elif datetime_range[0][-1] == 'h':
                    datetime_range_delta = {'hours': int(datetime_range[0][:-1])}
                elif datetime_range[0][-1] == 'w':
                    datetime_range_delta = {'weeks': int(datetime_range[0][:-1])}
                attr_val = (
                    datetime.datetime.utcnow()
                    + datetime.timedelta(**datetime_range_delta)
                ).isoformat()
            else:
                if datetime_range[0][0] not in ['+', '-']:
                    attr_val = datetime_range[0]
                else:
                    attr_val = (
                        datetime.datetime.fromisoformat(datetime_range[1])
                        - datetime.timedelta(days=1)
                    ).isoformat()
        else:
            attr_val = datetime.datetime.utcnow().isoformat()
        return attr_val

    elif attr_type._type == 'KV_DICT':
        attr_val = {}
        if attr_type._args['req']:
            attr_val = {
                generate_attr(attr_type=ATTR.LITERAL(literal=[req])): generate_attr(
                    attr_type=attr_type._args['val']
                )
                for req in attr_type._args['req']
            }
        for _ in range(attr_type._args['min'] or 0):
            attr_val[generate_attr(attr_type=attr_type._args['key'])] = generate_attr(
                attr_type=attr_type._args['val']
            )
        if len(attr_val.keys()) < (attr_type._args['min'] or 0):
            attr_val = generate_attr(attr_type=attr_type)
        return attr_val

    elif attr_type._type == 'TYPED_DICT':
        attr_val = {
            child_attr: generate_attr(attr_type=attr_type._args['dict'][child_attr])
            for child_attr in attr_type._args['dict'].keys()
        }
        return attr_val

    elif attr_type._type == 'EMAIL':
        attr_val = f'some-{math.ceil(random.random() * 10000)}@mail.provider.com'
        if attr_type._args['allowed_domains']:
            if attr_type._args['strict']:
                domain = 'mail.provider.com'
            else:
                domain = 'provider.com'
            attr_val = attr_val.replace(
                domain, random.choice(attr_type._args['allowed_domains'])
            )
        return attr_val

    elif attr_type._type == 'FILE':
        attr_file_type = 'text/plain'
        attr_file_extension = 'txt'
        if attr_type._args['types']:
            for file_type in attr_type._args['types']:
                if '/' in file_type:
                    attr_file_type = file_type
                if '*.' in file_type:
                    attr_file_extension = file_type.replace('*.', '')
        file_name = f'__file-{math.ceil(random.random() * 10000)}.{attr_file_extension}'
        return {
            'name': file_name,
            'lastModified': 100000,
            'type': attr_file_type,
            'size': 6,
            'content': b'__file',
        }

    elif attr_type._type == 'FLOAT':
        if attr_type._args['ranges']:
            attr_val = random.choice(
                range(
                    math.ceil(attr_type._args['ranges'][0][0]),
                    math.floor(attr_type._args['ranges'][0][1]),
                )
            )
            if (
                attr_val != attr_type._args['ranges'][0][0]
                and (attr_val - 0.01) != attr_type._args['ranges'][0][0]
            ):
                attr_val -= 0.01
            elif (attr_val + 0.01) < attr_type._args['ranges'][0][1]:
                attr_val += 0.01
            else:
                attr_val = float(attr_val)
        else:
            attr_val = random.random() * 10000
        return attr_val

    elif attr_type._type == 'GEO':
        return {
            'type': 'Point',
            'coordinates': [
                math.ceil(random.random() * 100000) / 1000,
                math.ceil(random.random() * 100000) / 1000,
            ],
        }

    elif attr_type._type == 'ID':
        return ObjectId()

    elif attr_type._type == 'INT':
        if attr_type._args['ranges']:
            attr_val = random.choice(
                range(attr_type._args['ranges'][0][0], attr_type._args['ranges'][0][1])
            )
        else:
            attr_val = math.ceil(random.random() * 10000)
        return attr_val

    elif attr_type._type == 'IP':
        return '127.0.0.1'

    elif attr_type._type == 'LIST':
        return [
            generate_attr(attr_type=random.choice(attr_type._args['list']))
            for _ in range(attr_type._args['min'] or 0)
        ]

    elif attr_type._type == 'LOCALE':
        return {
            locale: f'__locale-{math.ceil(random.random() * 10000)}'
            for locale in Config.locales
        }

    elif attr_type._type == 'LOCALES':
        return Config.locale

    elif attr_type._type == 'PHONE':
        attr_phone_code = '000'
        if attr_type._args['codes']:
            attr_phone_code = random.choice(attr_type._args['codes'])
        return f'+{attr_phone_code}{math.ceil(random.random() * 10000)}'

    elif attr_type._type == 'STR':
        if attr_type._args['pattern']:
            logger.warning(
                'Generator for Attr Type STR can\'t handle patterns. Ignoring.'
            )
        return f'__str-{math.ceil(random.random() * 10000)}'

    elif attr_type._type == 'TIME':
        if attr_type._args['ranges']:
            datetime_range = attr_type._args['ranges'][0]
            # [DOC] Be lazy! find a whether start, end of range is a datetime and base the value on it
            if datetime_range[0][0] in ['+', '-'] and datetime_range[1][0] in [
                '+',
                '-',
            ]:
                # [DOC] Both start, end are dynamic, process start
                datetime_range_delta = {}
                if datetime_range[0][-1] == 's':
                    datetime_range_delta = {'seconds': int(datetime_range[0][:-1])}
                elif datetime_range[0][-1] == 'm':
                    datetime_range_delta = {'minutes': int(datetime_range[0][:-1])}
                elif datetime_range[0][-1] == 'h':
                    datetime_range_delta = {'hours': int(datetime_range[0][:-1])}
                attr_val = (
                    (
                        datetime.datetime.utcnow()
                        + datetime.timedelta(**datetime_range_delta)
                    )
                    .isoformat()
                    .split('T')[1]
                )
            else:
                if datetime_range[0][0] not in ['+', '-']:
                    attr_val = datetime_range[0]
                else:
                    # [REF]: https://stackoverflow.com/a/656394/2393762
                    attr_val = (
                        (
                            datetime.datetime.combine(
                                datetime.date.today(),
                                datetime.time.fromisoformat(datetime_range[1]),
                            )
                            - datetime.timedelta(minutes=1)
                        )
                        .isoformat()
                        .split('T')[1]
                    )
        else:
            attr_val = datetime.datetime.utcnow().isoformat().split('T')[1]
        return attr_val

    elif attr_type._type == 'URI_WEB':
        attr_val = f'https://sub.domain.com/page-{math.ceil(random.random() * 10000)}/'
        if attr_type._args['allowed_domains']:
            if attr_type._args['strict']:
                domain = 'sub.domain.com'
            else:
                domain = 'domain.com'
            attr_val = attr_val.replace(
                domain, random.choice(attr_type._args['allowed_domains'])
            )
        return attr_val

    elif attr_type._type == 'LITERAL':
        attr_val = random.choice(attr_type._args['literal'])
        return attr_val

    elif attr_type._type == 'UNION':
        attr_val = generate_attr(attr_type=random.choice(attr_type._args['union']))
        return attr_val

    raise Exception(f'Unknown generator attr \'{attr_type}\'')



from nawah.classes import ATTR
from nawah.enums import NAWAH_VALUES

from typing import Dict, Any

import copy


def encode_attr_type(*, attr_type: ATTR) -> Dict[str, Any]:
    encoded_attr_type: Dict[str, Any] = {
        'type': attr_type._type,
        'args': copy.deepcopy(attr_type._args),
        'allow_none': attr_type._default != NAWAH_VALUES.NONE_VALUE,
        'default': attr_type._default
        if attr_type._default != NAWAH_VALUES.NONE_VALUE
        else None,
    }
    # [DOC] Process args of type ATTR
    if attr_type._type == 'LIST':
        for i in range(len(attr_type._args['list'])):
            encoded_attr_type['args']['list'][i] = encode_attr_type(
                attr_type=attr_type._args['list'][i]
            )
    elif attr_type._type == 'TYPED_DICT':
        for dict_attr in attr_type._args['dict'].keys():
            encoded_attr_type['args']['dict'][dict_attr] = encode_attr_type(
                attr_type=attr_type._args['dict'][dict_attr]
            )
    elif attr_type._type == 'KV_DICT':
        encoded_attr_type['args']['key'] = encode_attr_type(
            attr_type=attr_type._args['key']
        )
        encoded_attr_type['args']['val'] = encode_attr_type(
            attr_type=attr_type._args['val']
        )
    elif attr_type._type == 'UNION':
        for i in range(len(attr_type._args['union'])):
            encoded_attr_type['args']['union'][i] = encode_attr_type(
                attr_type=attr_type._args['union'][i]
            )
    elif attr_type._type == 'TYPE':
        if callable(encoded_attr_type['args']['type']):
            raise Exception('Attr Type TYPE with callable \'type\' can\'t be encoded.')

        del encoded_attr_type['args']['func']

    return encoded_attr_type



from nawah.config import Config

import argparse

from ._launch import launch


def generate_ref(args: argparse.Namespace):
    # [DOC] Update Config with Nawah framework CLI args
    from nawah.config import Config

    Config.generate_ref = True
    launch(args=args, custom_launch='generate_ref')


def generate_models(args: argparse.Namespace):
    # [DOC] Update Config with Nawah framework CLI args
    from nawah.config import Config

    Config.generate_models = True
    launch(args=args, custom_launch='generate_models')



import datetime
import logging
import re
from typing import TYPE_CHECKING, Dict, List

from nawah.classes import ATTR
from nawah.config import Config

if TYPE_CHECKING:
    from nawah.base_module import BaseModule

logger = logging.getLogger('nawah')


def _generate_models():
    # [DOC] Initialise _api_models Config Attr
    Config._api_models = '// Nawah Models\n'
    # [DOC] Add interface for DOC, LOCALE, LOCALES, FILE typing
    Config._api_models += 'interface Doc { _id: string; };\n'
    Config._api_models += 'export interface LOCALE { '
    for locale in Config.locales:
        Config._api_models += locale
        if locale != Config.locale:
            Config._api_models += '?'
        Config._api_models += ': string; '
    Config._api_models += '};\n'
    Config._api_models += (
        'export type LOCALES = \'' + '\' | \''.join(Config.locales) + '\';\n'
    )
    Config._api_models += 'export type ID<T> = string & T;\n'
    Config._api_models += 'export interface FILE<T> { name: string; lastModified: number; type: T; size: number; content: string | boolean; };\n'
    # [DOC] Iterate over packages in ascending order
    for package in sorted(Config.modules_packages.keys()):
        # [DOC] Add package header
        Config._api_models += f'\n// Package: {package.replace("modules.", "")}\n'
        if not len(Config.modules_packages[package]):
            Config._api_models += f'// No modules\n'
        # [DOC] Iterate over package modules in ascending order
        for module in sorted(Config.modules_packages[package]):
            module_class = (
                str(Config.modules[module].__class__).split('.')[-1].split('\'')[0]
            )
            # [DOC] Add module header
            Config._api_models += f'// Module: {module_class}\n'

            # [DOC] Add module interface definition
            Config._api_models += (
                f'export interface {module_class} extends String, Doc {{\n'
            )
            # [DOC] Iterate over module attrs to add attrs types, defaults (if any)
            for attr in Config.modules[module].attrs.keys():
                attr_model = ''
                if Config.modules[module].attrs[attr]._desc:
                    attr_model += f'\t// @property {{__TYPE__}} {Config.modules[module].attrs[attr]._desc}\n'
                attr_model += f'\t{attr}__DEFAULT__: __TYPE__;\n'
                for default_attr in Config.modules[module].defaults.keys():
                    if (
                        default_attr == attr
                        or default_attr.startswith(f'{attr}.')
                        or default_attr.startswith(f'{attr}:')
                    ):
                        # [DOC] Attr is in defaults, indicate the same
                        attr_model = attr_model.replace('__DEFAULT__', '?')
                # [DOC] Attempt to replace __DEFAULT__ with empty string if it still exists, effectively no default value
                attr_model = attr_model.replace('__DEFAULT__', '')

                # [DOC] Add typing
                attr_model = attr_model.replace(
                    '__TYPE__',
                    _generate_model_typing(
                        module=Config.modules[module],
                        attr_name=attr,
                        attr_type=Config.modules[module].attrs[attr],
                    ),
                )

                Config._api_models += attr_model

            # [DOC] Add closing braces
            Config._api_models += '};\n'
        Config._api_models += '\n'
    import os

    models_file = os.path.join(
        Config._app_path,
        'models',
        f'NAWAH_API_MODELS_{datetime.datetime.utcnow().strftime("%d-%b-%Y")}.ts',
    )
    with open(models_file, 'w') as f:
        f.write(Config._api_models)
        logger.info(f'API models generated and saved to: \'{models_file}\'. Exiting.')
        exit(0)


def _generate_model_typing(*, module: 'BaseModule', attr_name: str, attr_type: ATTR):
    if attr_type._type == 'ANY':
        return 'any'

    elif attr_type._type == 'ACCESS':
        return '{ anon: boolean; users: Array<string>; groups: Array<string>; }'

    elif attr_type._type == 'BOOL':
        return 'boolean'

    elif attr_type._type == 'COUNTER':
        return 'string'

    elif attr_type._type == 'DATE':
        return 'string'

    elif attr_type._type == 'DATETIME':
        return 'string'

    elif attr_type._type == 'DYNAMIC_ATTR':
        return '{ type: string; args: { [key: string]: any; }; allow_none?: boolean; default: any; }'

    elif attr_type._type == 'DYNAMIC_VAL':
        return 'any'

    elif attr_type._type == 'KV_DICT':
        key_typing = _generate_model_typing(
            module=module, attr_name=attr_name, attr_type=attr_type._args['key']
        )
        val_typing = _generate_model_typing(
            module=module, attr_name=attr_name, attr_type=attr_type._args['val']
        )
        return f'{{ [key: {key_typing}]: {val_typing} }}'

    elif attr_type._type == 'TYPED_DICT':
        typing = '{ '
        for child_attr_type in attr_type._args['dict'].keys():
            typing += child_attr_type
            typing += ': '
            typing += _generate_model_typing(
                module=module,
                attr_name=attr_name,
                attr_type=attr_type._args['dict'][child_attr_type],
            )
            typing += '; '
        typing += '}'
        return typing

    elif attr_type._type == 'EMAIL':
        return 'string'

    elif attr_type._type == 'FILE':
        types = 'string'
        if attr_type._args['types']:
            types = '\'' + '\' | \''.join(attr_type._args['types']) + '\''
            if '*' in types:
                types = 'string'

        return f'FILE<{types}>'

    elif attr_type._type == 'FLOAT':
        return 'number'

    elif attr_type._type == 'GEO':
        return '{ type: \'Point\'; coordinates: [number, number]; }'

    elif attr_type._type == 'ID':
        for attr in module.extns.keys():
            # [TODO] Validate whether Extension Instructions as ATTR breaks this code
            if attr.split('.')[0].split(':')[0] == attr_name:
                # [REF]: https://dev.to/rrampage/snake-case-to-camel-case-and-back-using-regular-expressions-and-python-m9j

                extn_module_class = re.sub(
                    r'(.*?)_([a-zA-Z])',
                    lambda match: match.group(1) + match.group(2).upper(),
                    module.extns[attr].module,
                )
                return f'ID<{extn_module_class[0].upper()}{extn_module_class[1:]}>'
        return 'ID<string>'

    elif attr_type._type == 'INT':
        return 'number'

    elif attr_type._type == 'IP':
        return 'string'

    elif attr_type._type == 'LIST':
        list_typings = []
        for child_attr_type in attr_type._args['list']:
            list_typings.append(
                _generate_model_typing(
                    module=module, attr_name=attr_name, attr_type=child_attr_type
                )
            )
        list_typing = ' | '.join(list_typings)
        return f'Array<{list_typing}>'

    elif attr_type._type == 'LOCALE':
        return 'LOCALE'

    elif attr_type._type == 'LOCALES':
        return 'LOCALES'

    elif attr_type._type == 'PHONE':
        return 'string'

    elif attr_type._type == 'STR':
        return 'string'

    elif attr_type._type == 'TIME':
        return 'string'

    elif attr_type._type == 'URI_WEB':
        return 'string'

    elif attr_type._type == 'LITERAL':
        return '\'' + '\' | \''.join(attr_type._args['literal']) + '\''

    elif attr_type._type == 'UNION':
        return ' | '.join(
            [
                _generate_model_typing(
                    module=module, attr_name=attr_name, attr_type=child_attr_type
                )
                for child_attr_type in attr_type._args['union']
            ]
        )

    elif attr_type._type == 'TYPE':
        return 'any'



from nawah.classes import (
    Query,
    ATTR,
    InvalidCallArgsException,
    InvalidAttrException,
    ConvertAttrException,
    NAWAH_ENV,
)
from nawah.utils import validate_attr

from typing import Union, Dict, Any, List, Literal, Iterable, cast


async def _validate_args(
    *,
    args: Union[Query, Dict[str, Any]],
    args_list_label: str,
    args_list: List[Dict[str, ATTR]],
    env: NAWAH_ENV,
):
    if not args_list:
        return

    sets_check: List[Dict[str, Literal[True, 'missing', 'invalid', 'convert']]] = []

    for args_set in args_list:
        set_status = True
        set_check = len(sets_check)
        sets_check.append({arg: True for arg in args_set.keys()})

        if args_list_label == 'query':
            args_check: Iterable = args
        elif args_list_label == 'doc':
            args = cast(Dict[str, Any], args)
            args_check = args.keys()

        for arg in args_set.keys():
            if arg not in args_check:
                set_status = False
                sets_check[set_check][arg] = 'missing'
            else:
                try:
                    if args_list_label == 'query' and arg[0] != '$':
                        for i in range(len(args[arg])):
                            if ':' not in arg:
                                args[arg][i] = await validate_attr(
                                    mode='create',
                                    attr_name=arg,
                                    attr_type=args_set[arg],
                                    attr_val=args[arg][i],
                                    env=env,
                                )
                            else:
                                attr_val = args[arg][i]
                                if (arg_oper := arg.split(':')[1]) in ['*', '$eq']:
                                    if (
                                        type(args[arg][i])
                                        and arg_oper in args[arg][i].keys()
                                    ):
                                        attr_val = args[arg][i][arg_oper]
                                attr_val = await validate_attr(
                                    mode='create',
                                    attr_name=arg,
                                    attr_type=args_set[arg],
                                    attr_val=attr_val,
                                    env=env,
                                )
                                if arg_oper in ['*', '$eq']:
                                    args[arg][i] = attr_val
                                else:
                                    args[arg][i] = {arg_oper: attr_val}

                    elif args_list_label == 'query' and arg[0] == '$':
                        args[arg] = await validate_attr(
                            mode='create',
                            attr_name=arg,
                            attr_type=args_set[arg],
                            attr_val=args[arg],
                            env=env,
                        )
                    elif args_list_label == 'doc':
                        args[arg] = await validate_attr(
                            mode='create',
                            attr_name=arg,
                            attr_type=args_set[arg],
                            attr_val=args[arg],
                            env=env,
                        )
                except InvalidAttrException:
                    set_status = False
                    sets_check[set_check][arg] = 'invalid'
                except ConvertAttrException:
                    set_status = False
                    sets_check[set_check][arg] = 'convert'

        if set_status:
            return

    raise InvalidCallArgsException(sets_check)



import copy

from bson import ObjectId

from nawah.base_module import BaseModule
from nawah.classes import ATTR, METHOD, PERM
from nawah.config import Config
from nawah.enums import NAWAH_VALUES, Event
from nawah.registry import Registry
from nawah.utils import encode_attr_type, validate_attr


class User(BaseModule):
    '''`User` module provides data type and controller for users in Nawah eco-system. The permissions of the module methods are designed to be as secure for exposed calls, and as flexible for privileged-access.'''

    collection = 'users'
    attrs = {
        'name': ATTR.LOCALE(desc='Name of the user as `LOCALE`.'),
        'locale': ATTR.LOCALES(desc='Default locale of the user.'),
        'create_time': ATTR.DATETIME(
            desc='Python `datetime` ISO format of the doc creation.'
        ),
        'login_time': ATTR.DATETIME(
            desc='Python `datetime` ISO format of the last login.'
        ),
        'groups': ATTR.LIST(
            desc='List of `_id` for every group the user is member of.',
            list=[ATTR.ID(desc='`_id` of Group doc the user is member of.')],
        ),
        'privileges': ATTR.KV_DICT(
            desc='Privileges of the user. These privileges are always available to the user regardless of whether groups user is part of have them or not.',
            key=ATTR.STR(),
            val=ATTR.LIST(list=[ATTR.STR()]),
        ),
        'status': ATTR.LITERAL(
            desc='Status of the user to determine whether user has access to the app or not.',
            literal=['active', 'banned', 'deleted', 'disabled_password'],
        ),
    }
    defaults = {'login_time': None, 'status': 'active', 'groups': [], 'privileges': {}}
    unique_attrs = []
    methods = {
        'read': METHOD(
            permissions=[
                PERM(privilege='admin'),
                PERM(privilege='read', query_mod={'_id': '$__user'}),
            ]
        ),
        'create': METHOD(permissions=[PERM(privilege='admin')]),
        'update': METHOD(
            permissions=[
                PERM(privilege='admin', doc_mod={'groups': None}),
                PERM(
                    privilege='update',
                    query_mod={'_id': '$__user'},
                    doc_mod={'groups': None, 'privileges': None},
                ),
            ],
            query_args={'_id': ATTR.ID()},
        ),
        'delete': METHOD(
            permissions=[
                PERM(privilege='admin'),
                PERM(privilege='delete', query_mod={'_id': '$__user'}),
            ],
            query_args={'_id': ATTR.ID()},
        ),
        'read_privileges': METHOD(
            permissions=[
                PERM(privilege='admin'),
                PERM(privilege='read', query_mod={'_id': '$__user'}),
            ],
            query_args={'_id': ATTR.ID()},
        ),
        'add_group': METHOD(
            permissions=[PERM(privilege='admin')],
            query_args={'_id': ATTR.ID()},
            doc_args=[{'group': ATTR.ID()}, {'group': ATTR.LIST(list=[ATTR.ID()])}],
        ),
        'delete_group': METHOD(
            permissions=[PERM(privilege='admin')],
            query_args={'_id': ATTR.ID(), 'group': ATTR.ID()},
        ),
        'retrieve_file': METHOD(permissions=[PERM(privilege='__sys')], get_method=True),
        'create_file': METHOD(permissions=[PERM(privilege='__sys')]),
        'delete_file': METHOD(permissions=[PERM(privilege='__sys')]),
    }

    async def on_read(self, results, skip_events, env, query, doc, payload):
        for i in range(len(results['docs'])):
            user = results['docs'][i]
            for auth_attr in Config.user_attrs.keys():
                del user[f'{auth_attr}_hash']
            if len(Config.user_doc_settings):
                setting_results = await Registry.module('setting').read(
                    skip_events=[Event.PERM, Event.ARGS],
                    env=env,
                    query=[
                        {'user': user['_id'], 'var': {'$in': Config.user_doc_settings}}
                    ],
                )
                user_doc_settings = copy.copy(Config.user_doc_settings)
                if setting_results['args']['count']:
                    for setting_doc in setting_results['args']['docs']:
                        user_doc_settings.remove(setting_doc['var'])
                        user[setting_doc['var']] = setting_doc['val']
                # [DOC] Forward-compatibility: If user was created before presence of any user_doc_settings, add them with default value
                for setting_attr in user_doc_settings:
                    user[setting_attr] = Config.user_settings[setting_attr].default
                    # [DOC] Set NAWAH_VALUES.NONE_VALUE to None if it was default
                    if user[setting_attr] == NAWAH_VALUES.NONE_VALUE:
                        user[setting_attr] = None
        return (results, skip_events, env, query, doc, payload)

    async def pre_create(self, skip_events, env, query, doc, payload):
        if Event.ARGS not in skip_events:
            doc['groups'] = [ObjectId('f00000000000000000000013')]
        user_settings = {}
        for attr in Config.user_settings.keys():
            if Config.user_settings[attr].type == 'user_sys':
                user_settings[attr] = copy.deepcopy(Config.user_settings[attr].default)
            else:
                if attr in doc.keys():
                    try:
                        await validate_attr(
                            mode='create',
                            attr_name=attr,
                            attr_type=Config.user_settings[attr].val_type,
                            attr_val=doc[attr],
                        )
                        user_settings[attr] = doc[attr]
                    except:
                        raise self.exception(
                            status=400,
                            msg=f'Invalid settings attr \'{attr}\' for \'create\' request on module \'CORE_USER\'',
                            args={'code': 'INVALID_ATTR'},
                        )

                else:
                    if Config.user_settings[attr].default == NAWAH_VALUES.NONE_VALUE:
                        raise self.exception(
                            status=400,
                            msg=f'Missing settings attr \'{attr}\' for \'create\' request on module \'CORE_USER\'',
                            args={'code': 'MISSING_ATTR'},
                        )

                    else:
                        user_settings[attr] = copy.deepcopy(
                            Config.user_settings[attr].default
                        )
        payload['user_settings'] = user_settings
        return (skip_events, env, query, doc, payload)

    async def on_create(self, results, skip_events, env, query, doc, payload):
        if 'user_settings' in payload.keys():
            for setting in payload['user_settings'].keys():
                setting_results = await Registry.module('setting').create(
                    skip_events=[Event.PERM, Event.ARGS],
                    env=env,
                    doc={
                        'user': results['docs'][0]['_id'],
                        'var': setting,
                        'val_type': encode_attr_type(
                            attr_type=Config.user_settings[setting].val_type
                        ),
                        'val': payload['user_settings'][setting],
                        'type': Config.user_settings[setting].type,
                    },
                )
                if setting_results['status'] != 200:
                    return setting_results
        return (results, skip_events, env, query, doc, payload)

    async def read_privileges(self, env, query):
        # [DOC] Confirm _id is valid
        results = await self.read(
            skip_events=[Event.PERM], env=env, query=[{'_id': query['_id'][0]}]
        )
        if not results['args']['count']:
            raise self.exception(
                status=400, msg='User is invalid.', args={'code': 'INVALID_USER'}
            )
        user = results['args']['docs'][0]
        for group in user['groups']:
            group_results = await Registry.module('group').read(
                skip_events=[Event.PERM], env=env, query=[{'_id': group}]
            )
            group = group_results['args']['docs'][0]
            for privilege in group['privileges'].keys():
                if privilege not in user['privileges'].keys():
                    user['privileges'][privilege] = []
                for i in range(len(group['privileges'][privilege])):
                    if (
                        group['privileges'][privilege][i]
                        not in user['privileges'][privilege]
                    ):
                        user['privileges'][privilege].append(
                            group['privileges'][privilege][i]
                        )
        return results

    async def add_group(self, skip_events, env, query, doc):
        # [DOC] Check for list group attr
        if type(doc['group']) == list:
            for i in range(0, len(doc['group']) - 1):
                await self.add_group(
                    skip_events=skip_events,
                    env=env,
                    query=query,
                    doc={'group': doc['group'][i]},
                )
            doc['group'] = doc['group'][-1]
        # [DOC] Confirm all basic args are provided
        doc['group'] = ObjectId(doc['group'])
        # [DOC] Confirm group is valid
        results = await Registry.module('group').read(
            skip_events=[Event.PERM], env=env, query=[{'_id': doc['group']}]
        )
        if not results['args']['count']:
            raise self.exception(
                status=400, msg='Group is invalid.', args={'code': 'INVALID_GROUP'}
            )
        # [DOC] Get user details
        results = await self.read(skip_events=[Event.PERM], env=env, query=query)
        if not results['args']['count']:
            raise self.exception(
                status=400, msg='User is invalid.', args={'code': 'INVALID_USER'}
            )
        user = results['args']['docs'][0]
        # [DOC] Confirm group was not added before
        if doc['group'] in user['groups']:
            raise self.exception(
                status=400,
                msg='User is already a member of the group.',
                args={'code': 'GROUP_ADDED'},
            )
        user['groups'].append(doc['group'])
        # [DOC] Update the user
        results = await self.update(
            skip_events=[Event.PERM],
            env=env,
            query=query,
            doc={'groups': user['groups']},
        )
        # [DOC] if update fails, return update results
        if results['status'] != 200:
            return results
        # [DOC] Check if the updated User doc belongs to current session and update it
        if env['session']['user']['_id'] == user['_id']:
            user_results = await Registry.module('user').read_privileges(
                skip_events=[Event.PERM], env=env, query=[{'_id': user['_id']}]
            )
            env['session']['user'] = user_results['args']['docs'][0]

        return results

    async def delete_group(self, env, query):
        # [DOC] Confirm group is valid
        results = await Registry.module('group').read(
            skip_events=[Event.PERM], env=env, query=[{'_id': query['group'][0]}]
        )
        if not results['args']['count']:
            raise self.exception(
                status=400, msg='Group is invalid.', args={'code': 'INVALID_GROUP'}
            )
        # [DOC] Get user details
        results = await self.read(
            skip_events=[Event.PERM], env=env, query=[{'_id': query['_id'][0]}]
        )
        if not results['args']['count']:
            raise self.exception(
                status=400, msg='User is invalid.', args={'code': 'INVALID_USER'}
            )
        user = results['args']['docs'][0]
        # [DOC] Confirm group was not added before
        if query['group'][0] not in user['groups']:
            raise self.exception(
                status=400,
                msg='User is not a member of the group.',
                args={'code': 'GROUP_NOT_ADDED'},
            )
        # [DOC] Update the user
        results = await self.update(
            skip_events=[Event.PERM],
            env=env,
            query=[{'_id': query['_id'][0]}],
            doc={'groups': {'$del_val': [query['group'][0]]}},
        )
        # [DOC] if update fails, return update results
        if results['status'] != 200:
            return results
        # [DOC] Check if the updated User doc belongs to current session and update it
        if env['session']['user']['_id'] == user['_id']:
            user_results = await Registry.module('user').read_privileges(
                skip_events=[Event.PERM], env=env, query=[{'_id': user['_id']}]
            )
            env['session']['user'] = user_results['args']['docs'][0]

        return results



from nawah import __version__

from typing import Dict, Any

import argparse, logging, os, sys, re, json, tarfile, tempfile, urllib.request, subprocess, random, string

logger = logging.getLogger('nawah')
__api_level__ = '.'.join(__version__.split('.')[:-1])


def create(args: argparse.Namespace):
    global os, subprocess

    if args.app_name == 'nawah_app':
        logger.error(
            'Value for \'app_name\' CLI Arg is invalid. Name can\'t be \'nawah_app\''
        )
        exit(1)
    elif not re.match(r'^[a-z][a-z0-9_]+$', args.app_name):
        logger.error(
            'Value for \'app_name\' CLI Arg is invalid. Name should have only small letters, numbers, and underscores.'
        )
        exit(1)

    app_path = os.path.realpath(os.path.join(args.app_path, args.app_name))
    progress_path = os.path.realpath(
        os.path.join(args.app_path, args.app_name, 'progress.json')
    )

    progress = None

    if os.path.exists(app_path):
        logger.info(
            'Specified \'app_name\' already existing in \'app_path\'. Attempting to check for earlier progress.'
        )
        if os.path.exists(progress_path):
            logger.info('File \'progress.json\' found. Attempting to process it.')
            try:
                with open(progress_path, 'r') as progress_file:
                    progress_config = json.loads(progress_file.read())
                    progress = progress_config['step']
                    app_config = progress_config['config']
            except Exception as e:
                logger.error(
                    'An exception occurred while attempting to process file \'progress.json\'.'
                )
                logger.error(f'Exception details: {e}')
                logger.error('Exiting.')
                exit(1)
        else:
            logger.error('File \'progress.json\' was not found. Exiting.')
            exit(1)

    # [DOC] Populating app_config
    if not progress:
        app_config = _create_step_config(args=args)
        logger.info('This will create an app with the following config:')
        for config_attr, config_set in app_config.items():
            logger.info(f'- {config_attr}: \'{config_set[1]}\'')
    else:
        logger.info('Continuing to create app with loaded progress config:')
        for config_attr, config_set in app_config.items():
            logger.info(f'- {config_attr}: \'{config_set[1]}\'')

    # [DOC] Create app workspace
    if not progress:
        cli_dir, cli_filename = os.path.split(__file__)
        template_archive = os.path.join(cli_dir, 'template.tar.gz')

        logger.info(f'Attempting to extract template archive to: {app_path}')
        with tarfile.open(name=template_archive, mode='r:gz') as archive:
            archive.extractall(path=app_path)
        logger.info('Template archive extracted successfully!')

    if not progress or progress == 1:
        progress = None
        logger.info('Attempting to config app template for new Nawah app.')
        try:
            with open(
                os.path.realpath(
                    os.path.join(args.app_path, args.app_name, 'nawah_app.py')
                ),
                'r',
            ) as f:
                nawah_app_file = f.read()
            with open(
                os.path.realpath(
                    os.path.join(args.app_path, args.app_name, 'nawah_app.py')
                ),
                'w',
            ) as f:
                nawah_app_file = nawah_app_file.replace(
                    '__PROJECT_NAME__', args.app_name, 2
                )
                for config_set in app_config.values():
                    nawah_app_file = nawah_app_file.replace(
                        config_set[0], config_set[1], 1
                    )
                f.write(nawah_app_file)
        except Exception as e:
            _dump_progress(
                args=args, app_config=app_config, progress_path=progress_path, step=3
            )
            logger.error('An exception occurred.')
            logger.error(f'Exception details: {e}')
            logger.error('Exiting.')
            exit(1)

    if not progress or progress == 2:
        progress = None
        try:
            gitignore_path = os.path.realpath(
                os.path.join(args.app_path, args.app_name, '.gitignore')
            )
            with open(gitignore_path, 'r') as f:
                gitignore_file = f.read()
            with open(gitignore_path, 'w') as f:
                f.write(gitignore_file.replace('PROJECT_NAME', args.app_name, 1))
        except Exception as e:
            _dump_progress(
                args=args, app_config=app_config, progress_path=progress_path, step=4
            )
            logger.error('An exception occurred.')
            logger.error(f'Exception details: {e}')
            logger.error('Exiting.')
            exit(1)

    if not progress or progress == 3:
        progress = None
        try:
            readme_path = os.path.realpath(
                os.path.join(args.app_path, args.app_name, 'README.md')
            )
            with open(readme_path, 'w') as f:
                f.write(
                    f'''# {args.app_name}
This Nawas app project was created with Nawah CLI v{__version__}, with API Level {__api_level__}.'''
                )
        except Exception as e:
            _dump_progress(
                args=args, app_config=app_config, progress_path=progress_path, step=6
            )
            logger.error('An exception occurred.')
            logger.error(f'Exception details: {e}')
            logger.error('Exiting.')
            exit(1)

    if not progress or progress == 4:
        progress = None
        try:
            os.rename(
                os.path.realpath(
                    os.path.join(
                        args.app_path, args.app_name, 'packages', 'PROJECT_NAME'
                    )
                ),
                os.path.realpath(
                    os.path.join(
                        args.app_path, args.app_name, 'packages', args.app_name
                    )
                ),
            )
        except Exception as e:
            _dump_progress(
                args=args, app_config=app_config, progress_path=progress_path, step=7
            )
            logger.error('An exception occurred.')
            logger.error(f'Exception details: {e}')
            logger.error('Exiting.')
            exit(1)

    if not progress or progress == 5:
        progress = None
        try:
            test_integration_path = os.path.realpath(
                os.path.join(
                    args.app_path,
                    args.app_name,
                    'tests',
                    'integration',
                    'test_integration.py',
                )
            )
            with open(test_integration_path, 'r') as f:
                test_integration_file = f.read()
            with open(test_integration_path, 'w') as f:
                f.write(
                    test_integration_file.replace(
                        'PROJECT_NAME', args.app_name.upper(), 1
                    )
                )
        except Exception as e:
            _dump_progress(
                args=args, app_config=app_config, progress_path=progress_path, step=4
            )
            logger.error('An exception occurred.')
            logger.error(f'Exception details: {e}')
            logger.error('Exiting.')
            exit(1)

    if not progress or progress == 6:
        progress = None
        try:
            for unit_test_filename in ['test_read.py', 'test_create.py']:
                test_unit_path = os.path.realpath(
                    os.path.join(
                        args.app_path,
                        args.app_name,
                        'tests',
                        'unit',
                        'doc',
                        unit_test_filename,
                    )
                )
                with open(test_unit_path, 'r') as f:
                    test_unit_file = f.read()
                with open(test_unit_path, 'w') as f:
                    f.write(test_unit_file.replace('PROJECT_NAME', args.app_name, 1))
        except Exception as e:
            _dump_progress(
                args=args, app_config=app_config, progress_path=progress_path, step=4
            )
            logger.error('An exception occurred.')
            logger.error(f'Exception details: {e}')
            logger.error('Exiting.')
            exit(1)

    logger.info(f'Congrats! Your Nawah app {args.app_name} is successfully created!')


def _create_step_config(*, args: argparse.Namespace):
    app_config = {
        # [REF]: https://stackoverflow.com/a/47073723/2393762
        'admin_password': [
            '__ADMIN_PASSWORD__',
            ''.join(
                [
                    random.choice(
                        string.ascii_letters + string.digits + string.punctuation
                    )
                    for n in range(18)
                ]
            )
            .replace('\'', '^')
            .replace('\\', '/'),
        ],
        'anon_token_suffix': [
            '__ANON_TOKEN_SUFFIX__',
            ''.join([random.choice(string.digits) for n in range(24)]),
        ],
    }

    envs_defaults = {
        'dev_local': 'mongodb://localhost',
        'dev_server': 'mongodb://admin:admin@mongodb',
        'prod': 'mongodb://admin:admin@prod',
    }

    if not args.default_config:
        # [DOC] Allow user to specify custom config per the following Config Attrs
        logger.info(
            'Not detected \'default-config\' CLI Arg. Attempting to create Nawah app with custom Config.'
        )
        logger.info(
            'If you would like to have app created with default config, stop the process, and re-run Nawah CLI with \'default-config\' CLI Arg.'
        )
        # [DOC] envs: data_server
        logger.info(
            '\'envs\' Config Attr provides environment-specific configuration. You will be required to specify \'data_server\' for each of the default available Nawah environments, namely \'dev_local\', \'dev_server\', and \'prod\'.'
        )
        for env in envs_defaults.keys():
            while True:
                config_attr_val = input(
                    f'\n> What would be the value for \'data_server\' Config Attr for environment \'{env}\'; The connection string to connect to your MongoDB host? [{envs_defaults[env]}]\n- '
                )
                if not config_attr_val:
                    logger.info(
                        f'Setting \'data_server\' Config Attr for environment \'{env}\' to default \'{envs_defaults[env]}\'.'
                    )
                    app_config[f'{env}:data_server'] = [
                        f'__{env.upper()}_DATA_SERVER__',
                        (config_attr_val := envs_defaults[env]),
                    ]
                else:
                    logger.info(
                        f'Setting \'data_server\' Config Attr for environment \'{env}\' to: \'{config_attr_val}\'.'
                    )
                    app_config[f'{env}:data_server'] = [
                        f'__{env.upper()}_DATA_SERVER__',
                        config_attr_val,
                    ]
                break
        # [DOC] env
        while True:
            config_attr_val = input(
                '\n> What would be the value for \'env\'; Config Attr for default environnement to be used when invoking Nawah CLI \'launch\' command? [$__env.ENV]\n- '
            )
            if not config_attr_val:
                logger.info('Setting \'env\' Config Attr to default: \'$__env.ENV\'.')
                app_config['data_name'] = ['__ENV__', (config_attr_val := '$__env.ENV')]
                break
            elif config_attr_val not in list(envs_defaults.keys()) and not re.match(
                r'^\$__env\.[A-Za-z_]+$', config_attr_val
            ):
                logger.error(
                    '\'env\' Config Attr can only be one of the environments names defined in \'envs\' Config Attr, or a valid Env Variable.'
                )
            else:
                logger.info(f'Setting \'env\' Config Attr to: \'{config_attr_val}\'.')
                app_config['env'] = ['__ENV__', config_attr_val]
                break
        # [DOC] data_name
        while True:
            config_attr_val = input(
                '\n> What would be the value for \'data_name\'; Config Attr for database name to be created on \'data_server\'? [nawah_data]\n- '
            )
            if not config_attr_val:
                logger.info(
                    'Setting \'data_name\' Config Attr to default: \'nawah_data\'.'
                )
                app_config['data_name'] = [
                    '__DATA_NAME__',
                    (config_attr_val := 'nawah_data'),
                ]
                break
            elif not re.match(r'^[a-zA-Z0-9\-_]+$', config_attr_val):
                logger.error(
                    '\'data_name\' Config Attr can\'t have special characters other than underscores and hyphens in it.'
                )
            else:
                logger.info(
                    f'Setting \'data_name\' Config Attr to: \'{config_attr_val}\'.'
                )
                app_config['data_name'] = ['__DATA_NAME__', config_attr_val]
                break
        # [DOC] locales
        while True:
            config_attr_val = input(
                '\n> What would be the comma-separated, language_COUNTRY-formatted value for \'locales\'; Config Attr for localisations of your app? [ar_AE, en_AE]\n- '
            )
            if not config_attr_val:
                logger.info(
                    'Setting \'locales\' Config Attr to default: \'ar_AE, en_AE\'.'
                )
                locales = ['ar_AE', 'en_AE']
                app_config['locales'] = [
                    '__LOCALES__',
                    (config_attr_val := 'ar_AE\', \'en_AE'),
                ]
                break
            else:
                try:
                    locales = [
                        re.match(r'^([a-z]{2}_[A-Z]{2})$', locale.strip()).group(0)  # type: ignore
                        for locale in config_attr_val.split(',')
                    ]
                    logger.info(
                        f'Setting \'locales\' Config Attr to: \'{config_attr_val}\'.'
                    )
                    app_config['locales'] = ['__LOCALES__', '\', \''.join(locales)]
                    break
                except:
                    logger.error(
                        'An exception occurred while attempting to process value provided.'
                    )
                    logger.error(
                        '\'locales\' Config Attr value should be provided as comma-separated, language_COUNTRY-formatted list of localisations.'
                    )
        # [DOC] locale
        while True:
            config_attr_val = input(
                '\n> What would be the value for \'locale\'; Config Attr for default localisation of your app? [first value of \'locales\' Config Attr]\n- '
            )
            if not config_attr_val:
                logger.info(
                    f'Setting \'locale\' Config Attr to first value of \'locales\' Config Attr: \'{locales[0]}\'.'
                )
                app_config['locale'] = ['__LOCALE__', (config_attr_val := locales[0])]
                break
            elif config_attr_val not in locales:
                logger.error(
                    '\'locale\' Config Attr can only be one of the localisations defined in \'locales\' Config Attr.'
                )
            else:
                logger.info(
                    f'Setting \'locale\' Config Attr to: \'{config_attr_val}\'.'
                )
                app_config['locale'] = ['__LOCALE__', config_attr_val]
                break
        # [DOC] admin_doc: email
        while True:
            config_attr_val = input(
                '\n> What would be the value for \'admin_doc\'.\'email\'; Config Attr? [admin@app.nawah.localhost]\n- '
            )
            if not config_attr_val:
                logger.info(
                    'Setting \'admin_doc\'.\'email\' Config Attr to default: \'admin@app.nawah.localhost\'.'
                )
                app_config['admin_doc:email'] = [
                    '__ADMIN_DOC_EMAIL__',
                    (config_attr_val := 'admin@app.nawah.localhost'),
                ]
                break
            elif not re.match(r'^[^@]+@[^@]+\.[^@]+$', config_attr_val):
                logger.error(
                    '\'admin_doc\'.\'email\' Config Attr value is not a valid email address.'
                )
            else:
                logger.info(
                    f'Setting \'admin_doc\'.\'email\' Config Attr to: \'{config_attr_val}\'.'
                )
                app_config['admin_doc:email'] = ['__ADMIN_DOC_EMAIL__', config_attr_val]
                break
    else:
        for env in envs_defaults.keys():
            app_config[f'{env}:data_server'] = [
                f'__{env.upper()}_DATA_SERVER__',
                envs_defaults[env],
            ]
        app_config.update(
            {
                'env': ['__ENV__', '$__env.ENV'],
                'data_name': ['__DATA_NAME__', 'nawah_data'],
                'locales': ['__LOCALES__', '\', \''.join(['ar_AE', 'en_AE'])],
                'locale': ['__LOCALE__', 'ar_AE'],
                'admin_doc:email': ['__ADMIN_DOC_EMAIL__', 'admin@app.nawah.localhost'],
            }
        )

    return app_config


def _dump_progress(
    *,
    args: argparse.Namespace,
    app_config: Dict[Any, Any],
    progress_path: str,
    step: int,
):
    with open(progress_path, 'w') as progress_file:
        progress_file.write(
            json.dumps(
                {
                    'step': step,
                    'args': {
                        'app_path': args.app_path,
                        'app_name': args.app_name,
                    },
                    'config': app_config,
                }
            )
        )



"""Provides enums required for Nawah Fremework and apps"""

from enum import Enum, auto


class Event(Enum):
    """Identifies events in calls that can be skipped"""

    ARGS = auto()
    VALIDATE = auto()
    PERM = auto()
    CACHE = auto()
    PRE = auto()
    ON = auto()
    EXTN = auto()
    SOFT = auto()
    DIFF = auto()
    SYS_DOCS = auto()


class DELETE_STRATEGY(Enum):
    """Identifies strategies of delete calls"""

    SOFT_SKIP_SYS = auto()
    SOFT_SYS = auto()
    FORCE_SKIP_SYS = auto()
    FORCE_SYS = auto()


class LOCALE_STRATEGY(Enum):
    """Identifies strategies of values for attrs of Attr Type LOCALE"""

    DUPLICATE = auto()
    NONE_VALUE = auto()


class NAWAH_VALUES(Enum):
    """Identifies values to be used internally to pass secure values"""

    NONE_VALUE = auto()
    ALLOW_MOD = auto()



from dataclasses import dataclass
from typing import (TYPE_CHECKING, Any, Callable, Dict, List, Literal,
                    Optional, Protocol, TypedDict, Union, cast)

from bson import ObjectId
from croniter import croniter

from ._types import NAWAH_DOC

if TYPE_CHECKING:
    from ._attr import ATTR
    from ._types import NAWAH_ENV


CLIENT_APP = TypedDict(
    'CLIENT_APP',
    {
        'name': str,
        'type': Literal['web', 'ios', 'android'],
        'origin': List[str],
        'hash': str,
    },
)


class USER_SETTING:
    type: Literal['user', 'user_sys']
    val_type: 'ATTR'
    default: Optional[Any]

    def __init__(
        self,
        *,
        type: Literal['user', 'user_sys'],
        val_type: 'ATTR',
        default: Optional[Any] = None,
    ):
        self.type = type
        self.val_type = val_type
        self.default = default

    def _validate(self):
        from ._attr import ATTR

        # [DOC] Validate type
        if type(self.type) != str or self.type not in ['user', 'user_sys']:
            raise Exception(
                f'Invalid \'type\' of type \'{type(self.type)}\' with required \'user\', or \'user_sys\'.'
            )

        # [DOC] Validate val_type
        if type(self.val_type) != ATTR:
            raise Exception(
                f'Invalid \'val_type\' of type \'{type(self.val_type)}\' with required type \'ATTR\'.'
            )

        # [DOC] Validate val_type Attr Type
        ATTR.validate_type(attr_type=self.val_type)


ANALYTICS_EVENTS = TypedDict(
    'ANALYTICS_EVENTS',
    {
        'app_conn_verified': bool,
        'session_conn_auth': bool,
        'session_user_auth': bool,
        'session_conn_reauth': bool,
        'session_user_reauth': bool,
        'session_conn_deauth': bool,
        'session_user_deauth': bool,
    },
)


class SYS_DOC:
    module: str
    key: Optional[str]
    skip_args: bool
    doc: Optional[NAWAH_DOC]

    @property
    def key_value(self) -> Any:
        if not (self.doc):
            raise Exception(f'SYS_DOC instance was initialised with no \'doc\'.')
        self.key = cast(str, self.key)
        return self.doc[self.key]

    def __init__(
        self,
        *,
        module: str,
        key: str = None,
        skip_args: bool = False,
        doc: NAWAH_DOC = None,
    ):
        if doc != None:
            if type(doc) != dict:
                raise Exception(
                    f'Argument \'doc\' is not a valid \'NAWAH_DOC\'. Expecting type \'dict\' but got \'{type(dict)}\'.'
                )
            doc = cast(NAWAH_DOC, doc)
            if not key:
                key = '_id'
            if key not in doc.keys():
                raise Exception(f'Attr \'{key}\' is not present on \'doc\'.')
            if key == '_id' and type(doc["_id"]) != ObjectId:
                raise Exception(
                    f'Invalid attr \'_id\' of type \'{type(doc["_id"])}\' with required type \'ID\''
                )
        else:
            if key or skip_args:
                raise Exception(
                    'Arguments \'attr, skip_args\' should only be used with \'doc\'.'
                )

        self.module = module
        self.key = key
        self.skip_args = skip_args
        self.doc = doc


class JOB_CALLABLE(Protocol):
    def __call__(
        self,
        env: 'NAWAH_ENV',
    ) -> bool:
        ...


class JOB:
    job: JOB_CALLABLE
    schedule: str
    prevent_disable: bool
    _cron_schedule: croniter
    _next_time: Optional[str] = None
    _disabled: bool = False

    def __init__(
        self,
        *,
        job: JOB_CALLABLE,
        schedule: str,
        prevent_disable: bool = False,
    ):
        self.job = job
        self.schedule = schedule
        self.prevent_disable = prevent_disable


class L10N(dict):
    pass


@dataclass
class PACKAGE_CONFIG:
    api_level: Optional[str] = None
    version: Optional[str] = None
    emulate_test: Optional[bool] = None
    debug: Optional[bool] = None
    port: Optional[int] = None
    env: Optional[str] = None
    force_admin_check: Optional[bool] = None
    vars_types: Optional[Dict[str, 'ATTR']] = None
    vars: Optional[Dict[str, Any]] = None
    client_apps: Optional[Dict[str, CLIENT_APP]] = None
    analytics_events: Optional[ANALYTICS_EVENTS] = None
    conn_timeout: Optional[int] = None
    quota_anon_min: Optional[int] = None
    quota_auth_min: Optional[int] = None
    quota_ip_min: Optional[int] = None
    data_server: Optional[str] = None
    data_name: Optional[str] = None
    data_ssl: Optional[bool] = None
    data_ca_name: Optional[str] = None
    data_ca: Optional[str] = None
    data_disk_use: Optional[bool] = None
    data_azure_mongo: Optional[bool] = None
    cache_server: Optional[str] = None
    cache_db: Optional[int] = 0
    locales: Optional[List[str]] = None
    locale: Optional[str] = None
    admin_doc: Optional[NAWAH_DOC] = None
    admin_password: Optional[str] = None
    anon_token: Optional[str] = None
    anon_privileges: Optional[Dict[str, List[str]]] = None
    user_attrs: Optional[Dict[str, 'ATTR']] = None
    user_settings: Optional[Dict[str, USER_SETTING]] = None
    user_doc_settings: Optional[List[str]] = None
    groups: Optional[List[Dict[str, Any]]] = None
    default_privileges: Optional[Dict[str, List[str]]] = None
    data_indexes: Optional[List[Dict[str, Any]]] = None
    docs: Optional[List[SYS_DOC]] = None
    jobs: Optional[Dict[str, JOB]] = None
    gateways: Optional[Dict[str, Callable]] = None
    types: Optional[Dict[str, Callable]] = None


@dataclass
class APP_CONFIG(PACKAGE_CONFIG):
    name: Optional[str] = None
    version: Optional[str] = None
    default_package: Optional[str] = None
    debug: Optional[bool] = False
    port: Optional[int] = None
    env: Optional[str] = None
    envs: Optional[Dict[str, PACKAGE_CONFIG]] = None
    force_admin_check: Optional[bool] = None



import datetime
import json

from bson import ObjectId


class JSONEncoderStr(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, ObjectId):
            return str(o)

        if isinstance(o, datetime.datetime):
            return o.isoformat()

        if isinstance(o, bytes):
            return True

        try:
            return json.JSONEncoder.default(self, o)
        except TypeError:
            return str(o)


JSONEncoderStrIns = JSONEncoderStr()


class JSONEncoderDict(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, ObjectId):
            return {'$oid': str(o)}

        if isinstance(o, datetime.datetime):
            return o.isoformat()

        if isinstance(o, bytes):
            return True

        try:
            return json.JSONEncoder.default(self, o)
        except TypeError:
            return str(o)


JSONEncoderDictIns = JSONEncoderDict()


def JSONEncoder(object_id_dict=False) -> json.JSONEncoder:
    if object_id_dict:
        return JSONEncoderDictIns

    return JSONEncoderStrIns
